from __future__ import annotations

import jax.numpy as jnp
from jax import Array

from horqrux.matrices import OPERATIONS_DICT


def BitFlip(error_probability: float) -> tuple[Array, ...]:
    """
    Initialize the BitFlip gate.

    The bit flip channel is defined as:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho + p X \\rho X^{\\dagger}

    Args:
        error_probability (float): The probability of a bit flip error.

    Raises:
        ValueError: If the error_probability value is not a float.

    Returns:
        tuple[Array, ...]: Kraus operators for this protocol.
    """
    if (error_probability > 1.0) or (error_probability < 0.0):
        raise ValueError(f"The 'error_probability' value is incorrect. Got {error_probability}.")
    K0: Array = jnp.sqrt(1.0 - error_probability) * OPERATIONS_DICT["I"]
    K1: Array = jnp.sqrt(error_probability) * OPERATIONS_DICT["X"]
    kraus: tuple[Array, ...] = (K0, K1)
    return kraus


def PhaseFlip(error_probability: float) -> tuple[Array, ...]:
    """
    Initialize the PhaseFlip gate

    The phase flip channel is defined as:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho + p Z \\rho Z^{\\dagger}

    Args:
        error_probability (float): The probability of phase flip error.

    Raises:
        ValueError: If the error_probability value is not a float.

    Returns:
        tuple[Array, ...]: Kraus operators for this protocol.
    """

    if (error_probability > 1.0) or (error_probability < 0.0):
        raise ValueError("The error_probability value is not a correct probability")
    K0: Array = jnp.sqrt(1.0 - error_probability) * OPERATIONS_DICT["I"]
    K1: Array = jnp.sqrt(error_probability) * OPERATIONS_DICT["Z"]
    kraus: tuple[Array, ...] = (K0, K1)
    return kraus


def Depolarizing(error_probability: float) -> tuple[Array, ...]:
    """
    Initialize the Depolarizing gate.

    The depolarizing channel is defined as:

    .. math::
        \\rho \\Rightarrow (1-p) \\rho
            + p/3 X \\rho X^{\\dagger}
            + p/3 Y \\rho Y^{\\dagger}
            + p/3 Z \\rho Z^{\\dagger}

    Args:
        error_probability (float): The probability of phase flip error.

    Raises:
        ValueError: If the error_probability value is not a float.

    Returns:
        tuple[Array, ...]: Kraus operators for this protocol.
    """

    if (error_probability > 1.0) or (error_probability < 0.0):
        raise ValueError(f"The 'error_probability' value is incorrect. Got {error_probability}.")
    K0: Array = jnp.sqrt(1.0 - error_probability) * OPERATIONS_DICT["I"]
    K1: Array = jnp.sqrt(error_probability / 3.0) * OPERATIONS_DICT["X"]
    K2: Array = jnp.sqrt(error_probability / 3.0) * OPERATIONS_DICT["Y"]
    K3: Array = jnp.sqrt(error_probability / 3.0) * OPERATIONS_DICT["Z"]
    kraus: tuple[Array, ...] = (K0, K1, K2, K3)
    return kraus


def PauliChannel(error_probability: tuple[float, ...]) -> tuple[Array, ...]:
    """
    Initialize the PauliChannel gate.

    The pauli channel is defined as:

    .. math::
        \\rho \\Rightarrow (1-px-py-pz) \\rho
            + px X \\rho X^{\\dagger}
            + py Y \\rho Y^{\\dagger}
            + pz Z \\rho Z^{\\dagger}

    Args:
        error_probability (tuple[float, ...] | float): tuple containing probabilities
            of X, Y, and Z errors.

    Raises:
        ValueError: If the probabilities values do not sum up to 1.

    Returns:
        tuple[Array, ...]: Kraus operators for this protocol.
    """

    sum_prob = sum(error_probability)
    if sum_prob > 1.0:
        raise ValueError("The sum of probabilities can't be greater than 1.0")
    if any([probability > 1.0 or probability < 0.0 for probability in error_probability]):
        raise ValueError(f"The 'error_probability' values are incorrect. Got {error_probability}.")
    px, py, pz = (
        error_probability[0],
        error_probability[1],
        error_probability[2],
    )

    K0: Array = jnp.sqrt(1.0 - (px + py + pz)) * OPERATIONS_DICT["I"]
    K1: Array = jnp.sqrt(px) * OPERATIONS_DICT["X"]
    K2: Array = jnp.sqrt(py) * OPERATIONS_DICT["Y"]
    K3: Array = jnp.sqrt(pz) * OPERATIONS_DICT["Z"]
    kraus: tuple[Array, ...] = (K0, K1, K2, K3)
    return kraus


def AmplitudeDamping(error_probability: float) -> tuple[Array, ...]:
    """
    Initialize the AmplitudeDamping gate.

    The amplitude damping channel is defined as:

    .. math::
        \\rho \\Rightarrow K_0 \\rho K_0^{\\dagger} + K_1 \\rho K_1^{\\dagger}

    with:

    .. code-block:: python

        K0 = [[1, 0], [0, sqrt(1 - error_probability)]]
        K1 = [[0, sqrt(error_probability)], [0, 0]]

    Args:
        error_probability (float): The damping rate, indicating the probability of amplitude loss.

    Raises:
        ValueError: If the damping rate is not a correct probability.

    Returns:
        tuple[Array, ...]: Kraus operators for this protocol.
    """

    if (error_probability > 1.0) or (error_probability < 0.0):
        raise ValueError(f"The 'error_probability' value is incorrect. Got {error_probability}.")
    K0: Array = jnp.array([[1, 0], [0, jnp.sqrt(1 - error_probability)]], dtype=jnp.complex128)
    K1: Array = jnp.array([[0, jnp.sqrt(error_probability)], [0, 0]], dtype=jnp.complex128)
    kraus: tuple[Array, ...] = (K0, K1)
    return kraus


def PhaseDamping(error_probability: float) -> tuple[Array, ...]:
    """
    Initialize the PhaseDamping gate.

    The phase damping channel is defined as:

    .. math::
        \\rho \\Rightarrow K_0 \\rho K_0^{\\dagger} + K_1 \\rho K_1^{\\dagger}

     with:

    .. code-block:: python

        K0 = [[1, 0], [0, sqrt(1 - error_probability)]]
        K1 = [[0, 0], [0, sqrt(error_probability)]]

    Args:
        error_probability (float): The damping rate, indicating the probability of phase damping.

    Raises:
        ValueError: If the damping rate is not a correct probability.

    Returns:
        tuple[Array, ...]: Kraus operators for this protocol.
    """

    if (error_probability > 1.0) or (error_probability < 0.0):
        raise ValueError(f"The 'error_probability' value is incorrect. Got {error_probability}.")
    K0: Array = jnp.array([[1, 0], [0, jnp.sqrt(1 - error_probability)]], dtype=jnp.complex128)
    K1: Array = jnp.array([[0, 0], [0, jnp.sqrt(error_probability)]], dtype=jnp.complex128)
    kraus: tuple[Array, ...] = (K0, K1)
    return kraus


def GeneralizedAmplitudeDamping(error_probability: tuple[float, ...]) -> tuple[Array, ...]:
    """
    Initialize the GeneralizeAmplitudeDamping gate.

    The generalize amplitude damping channel is defined as:

    .. math::
        \\rho \\Rightarrow K_0 \\rho K_0^{\\dagger} + K_1 \\rho K_1^{\\dagger}
            + K_2 \\rho K_2^{\\dagger} + K_3 \\rho K_3^{\\dagger}

    with:

    .. code-block:: python

        K0 = sqrt(p) * [[1, 0], [0, sqrt(1 - rate)]]
        K1 = sqrt(p) * [[0, sqrt(rate)], [0, 0]]
        K2 = sqrt(1-p) * [[sqrt(1 - rate), 0], [0, 1]]
        K3 = sqrt(1-p) * [[0, 0], [sqrt(rate), 0]]

    Args:
        error_probability (tuple[float, ...] | float): The first float must be the probability
            of amplitude damping error, and the second float is the damping rate, indicating
            the probability of generalized amplitude damping.

    Raises:
        ValueError: If the damping rate is not a correct probability.

    Returns:
        tuple[Array, ...]: Kraus operators for this protocol.
    """

    probability = error_probability[0]
    rate = error_probability[1]
    if (probability > 1.0) or (probability < 0.0):
        raise ValueError(
            f"The first value of 'error_probability' value is incorrect. Got {probability}."
        )
    if (rate > 1.0) or (rate < 0.0):
        raise ValueError(f"The second value of 'error_probability' value is incorrect. Got {rate}.")

    K0: Array = jnp.sqrt(probability) * jnp.array(
        [[1, 0], [0, jnp.sqrt(1 - rate)]], dtype=jnp.complex128
    )
    K1: Array = jnp.sqrt(probability) * jnp.array(
        [[0, jnp.sqrt(rate)], [0, 0]], dtype=jnp.complex128
    )
    K2: Array = jnp.sqrt(1.0 - probability) * jnp.array(
        [[jnp.sqrt(1.0 - rate), 0], [0, 1]], dtype=jnp.complex128
    )
    K3: Array = jnp.sqrt(1.0 - probability) * jnp.array(
        [[0, 0], [jnp.sqrt(rate), 0]], dtype=jnp.complex128
    )
    kraus: tuple[Array, ...] = (K0, K1, K2, K3)
    return kraus



from __future__ import annotations

import jax
import jax.numpy as jnp
from jax._src.typing import DType


def default_complex_dtype() -> DType:
    if jax.config.jax_enable_x64:
        return jnp.complex128
    else:
        return jnp.complex64



from __future__ import annotations

from functools import partial
from operator import is_not
from typing import Any, Iterable
from uuid import uuid4

import jax.numpy as jnp
from jax import Array

from horqrux.primitives import Parametric, Primitive


def to_shift(gate: Parametric, shift_value: float) -> Any:
    """Create the shifted gate for PSR.

    Args:
        gate (Parametric): Gate to shift.
        shift_value (float): Shift value.

    Returns:
        Any: A new Parametric (Any type due to clsmethod)
    """
    children, aux_data = gate.tree_flatten()
    return Parametric.tree_unflatten(aux_data[:-1] + (aux_data[-1] + shift_value,), children)


def prepare_param_gates_seq(
    param_names: tuple[str, ...], gates: Iterable[Primitive]
) -> dict[str, tuple]:
    """Create a dictionary of parameter names and corresponding parametric gates.

    Args:
        param_names (tuple[str, ...]): Parameters.
        gates (Iterable[Primitive]): Sequence of gates.

    Returns:
        dict[str, tuple]: dictionary of parameter names and corresponding parametric gates.
    """
    param_to_gates: dict[str, tuple] = dict.fromkeys(param_names, tuple())
    for i, gate in enumerate(gates):
        if gate.is_parametric and gate.param in param_names:  # type: ignore[attr-defined]
            param_to_gates[gate.param] += (i,)  # type: ignore[attr-defined]
    return param_to_gates


def alter_gate_sequence(gates: Iterable[Any], ind_alter: int, shift_val: float) -> Any:
    """Create a sequence replacing the `ind_alter` gate by its shifted version.

    Args:
        gates (Iterable[Any]): sequence of gates.
        ind_alter (int): Index of gate to shift.
        shift_val (float): Shift value.

    Returns:
        Any: sequence of gates including shift.
    """
    gate_alter = gates[ind_alter]  # type: ignore[index]
    gate_shift = to_shift(gate_alter, shift_val)
    upper = min(ind_alter + 1, len(gates))  # type: ignore[arg-type]
    gates_seq = gates[:ind_alter] + [gate_shift] + gates[upper:]  # type: ignore[index]
    return gates_seq


def initialize_gpsr_ingredients(
    values: dict[str, float]
) -> tuple[tuple[str, ...], dict[str, float], Array]:
    """Initialize the parameter names, spectral_gap, and shift value for GPSR.

    Args:
        values (dict[str, float]): Parameter values.

    Returns:
        tuple[tuple[str, ...], dict[str, float], Array]: parameter names, spectral_gap, and shift value.
    """
    val_keys = tuple(values.keys())
    spectral_gap = dict.fromkeys(val_keys, 2.0)
    shift = jnp.pi / 2
    return val_keys, spectral_gap, shift


def extract_gate_names(gates: Iterable[Primitive]) -> list:
    """Extract gate names when a gate is parametric.

    Args:
        gates (Iterable[Primitive]): List of gates.

    Returns:
        list: list of names.
    """
    gate_names = list(
        map(
            lambda g: g.param if hasattr(g, "param") and isinstance(g.param, str) else None,
            gates,
        )
    )
    gate_names = list(filter(partial(is_not, None), gate_names))
    return gate_names


def spectral_gap_from_gates(
    param_to_gates_indices: dict[str, tuple], param_names: Iterable[str]
) -> dict[str, Array]:
    """Extract spectral gap from each gate.

    Only works when a parameter name is used by only one gate.

    Args:
        param_to_gates_indices (dict[str, tuple]): dictionary mapping
            parameter name to indices of gates.
        param_names (Iterable[str]): Name of parameters.

    Returns:
        dict[str, Array]: Parameter names mapped with the spectral gap.
    """
    return {param: param_to_gates_indices[param][0].spectral_gap for param in param_names}


def new_named_parametric(gate: Primitive) -> Parametric:
    """Create a new Parametric with a different name for GPSR with repeated parameters.

    Args:
        gate (Parametric): Gate input.

    Returns:
        Parametric: New gate with different name.
    """
    children, aux_data = gate.tree_flatten()
    new_name = str(uuid4())
    new_gate: Parametric = Parametric.tree_unflatten(
        aux_data[:-2]
        + (
            new_name,
            aux_data[-1],
        ),
        children,
    )
    return new_gate


def create_renamed_operators(
    gates: Iterable[Primitive], param_names: Iterable[str]
) -> Iterable[Primitive]:
    """Create new quantum operations with renamed parameters.

    This is for GPSR with repeated parameters.

    Args:
        gates (Iterable[Primitive]): Original gates.
        param_names (Iterable[str]): Original aprameter names.

    Returns:
        Iterable[Primitive]: New gates.
    """
    new_ops: list = []
    for gate in gates:
        if gate.is_parametric and gate.param in param_names:  # type: ignore[arg-type, attr-defined]
            new_ops += [new_named_parametric(gate)]
        else:
            new_ops += [gate]
    return new_ops



from __future__ import annotations

from functools import singledispatch
from typing import Any

from jax import Array
from jax.experimental.sparse import BCOO

from horqrux.circuit import QuantumCircuit
from horqrux.composite.compose import Scale
from horqrux.composite.sequence import OpSequence
from horqrux.primitives.parametric import Parametric
from horqrux.primitives.primitive import Primitive


@singledispatch
def to_sparse(operations: Any) -> Any:
    raise NotImplementedError(
        f"to_sparse is not implemented for this argument type: {type(operations)}"
    )


@singledispatch
def to_dense(operations: Any) -> Any:
    raise NotImplementedError(
        f"to_dense is not implemented for this argument type: {type(operations)}"
    )


@to_sparse.register
def _(operations: Array) -> BCOO:
    return BCOO.fromdense(operations)


@to_dense.register
def _(operations: BCOO) -> Array:
    return operations.todense()


@to_sparse.register
def _(operations: Primitive) -> Any:
    children, aux_data = operations.tree_flatten()
    return operations.tree_unflatten(aux_data[:-1] + (True,), children)


@to_dense.register
def _(operations: Primitive) -> Any:
    children, aux_data = operations.tree_flatten()
    return operations.tree_unflatten(aux_data[:-1] + (False,), children)


@to_sparse.register
def _(operations: Parametric) -> Any:
    children, aux_data = operations.tree_flatten()
    return operations.tree_unflatten(aux_data[:-3] + (True,) + aux_data[-2:], children)


@to_dense.register
def _(operations: Parametric) -> Any:
    children, aux_data = operations.tree_flatten()
    return operations.tree_unflatten(aux_data[:-3] + (False,) + aux_data[-2:], children)


@to_sparse.register
def _(operations: OpSequence) -> OpSequence:
    return type(operations)([to_sparse(op) for op in operations.operations])


@to_dense.register
def _(operations: OpSequence) -> OpSequence:
    return type(operations)([to_dense(op) for op in operations.operations])


@to_sparse.register
def _(operations: Scale) -> Scale:
    return Scale(to_sparse(operations.operations[0]), operations.parameter)


@to_dense.register
def _(operations: Scale) -> Scale:
    return Scale(to_dense(operations.operations[0]), operations.parameter)


@to_sparse.register
def _(operations: QuantumCircuit) -> QuantumCircuit:
    return QuantumCircuit(operations.n_qubits, [to_sparse(op) for op in operations.operations])


@to_dense.register
def _(operations: QuantumCircuit) -> QuantumCircuit:
    return QuantumCircuit(operations.n_qubits, [to_dense(op) for op in operations.operations])



from __future__ import annotations

from functools import singledispatch
from typing import Any

import jax.numpy as jnp
from jax import Array
from jax.experimental.sparse import BCOO


@singledispatch
def verify_arrays(a: Any, b: Any, atol: float = 1e-08) -> bool:
    raise NotImplementedError("verify_arrays is not implemented")


@verify_arrays.register
def _(a: Array, b: Array, atol: float = 1e-08) -> bool:
    return jnp.allclose(a, b, atol)


@verify_arrays.register
def _(a: BCOO, b: BCOO, atol: float = 1e-08) -> bool:
    return jnp.allclose(a.todense(), b.todense(), atol)



from __future__ import annotations

from dataclasses import dataclass

from jax import Array
from jax.scipy.linalg import expm
from jax.tree_util import register_pytree_node_class

from horqrux.primitives.primitive import Primitive, QubitSupport


@register_pytree_node_class
@dataclass
class _HamiltonianEvolution(Primitive):
    """
    A slim wrapper class which evolves a 'hamiltonian'
    given a 'time_evolution' parameter and applies it to 'state' psi by doing: matrixexp(-iHt)|psi>
    """

    generator_name: str
    target: QubitSupport
    control: QubitSupport
    shift: float = 0.0

    def __post_init__(self) -> None:
        super().__post_init__()

    def _unitary(self, values: dict[str, Array] = dict()) -> Array:
        # note: GPSR trick when the same param_name is used in many operations
        time_val = values["time_evolution"] + self.shift
        return expm(values["hamiltonian"] * (-1j * time_val))


def HamiltonianEvolution(
    target: QubitSupport, control: QubitSupport = (None,)
) -> _HamiltonianEvolution:
    return _HamiltonianEvolution("I", target, control)



from __future__ import annotations

from .analytical import (
    analytical_expectation,
    analytical_gpsr_bwd,
    analytical_gpsr_fwd,
    jitted_analytical_exp,
)
from .shots import finite_shots, finite_shots_fwd, finite_shots_gpsr_backward, jitted_finite_shots



from __future__ import annotations

import chex
import jax
import jax.numpy as jnp
from absl.testing import parameterized

from horqrux import expectation, random_state, run
from horqrux.circuit import QuantumCircuit
from horqrux.composite import Observable
from horqrux.differentiation.gpsr import analytical_gpsr_bwd, finite_shots_gpsr_backward
from horqrux.primitives.parametric import RX
from horqrux.primitives.primitive import Z
from horqrux.utils.conversion import to_sparse
from horqrux.utils.operator_utils import density_mat
from tests.utils import verify_arrays

N_QUBITS = 2
GPSR_ATOL = 0.0001
SHOTS_ATOL = 0.01
N_SHOTS = 100_000


class GPSRTest(chex.TestCase):
    @chex.variants(with_jit=True, without_jit=True)
    @parameterized.parameters(True, False)
    def test_shots(self, same_name: bool) -> None:
        param_name = "theta"
        if same_name:
            x = jax.random.uniform(jax.random.key(0), (1,))
            ops = [RX(param_name, 0), RX(param_name, 1)]

            def values_to_dict(x):
                return {param_name: x}

        else:
            x = jax.random.uniform(jax.random.key(0), (2,))
            param_names = [param_name, param_name + "2"]
            ops = [RX(param_names[0], 0), RX(param_names[1], 1)]

            def values_to_dict(x):
                return {param_names[0]: x[0], param_names[1]: x[1]}

        circuit = QuantumCircuit(2, ops)
        observables = [Observable([Z(0)]), Observable([Z(1)])]
        state = random_state(N_QUBITS)

        @self.variant
        def exact(x):
            values = values_to_dict(x)
            return expectation(state, circuit, observables, values, diff_mode="ad")

        @self.variant
        def exact_dm(x):
            values = values_to_dict(x)
            return expectation(density_mat(state), circuit, observables, values, diff_mode="ad")

        @self.variant
        def exact_gpsr(x):
            values = values_to_dict(x)
            return expectation(state, circuit, observables, values, diff_mode="gpsr")

        @self.variant
        def exact_gpsr_dm(x):
            values = values_to_dict(x)
            return expectation(density_mat(state), circuit, observables, values, diff_mode="gpsr")

        @self.variant
        def shots(x):
            values = values_to_dict(x)
            return expectation(
                state,
                circuit,
                observables,
                values,
                diff_mode="gpsr",
                n_shots=N_SHOTS,
            )

        @self.variant
        def shots_dm(x):
            values = values_to_dict(x)
            return expectation(
                density_mat(state),
                circuit,
                observables,
                values,
                diff_mode="gpsr",
                n_shots=N_SHOTS,
            )

        expected_dm = density_mat(run(circuit, state, values_to_dict(x)))
        output_dm = run(circuit, density_mat(state), values_to_dict(x))
        assert jnp.allclose(expected_dm.array, output_dm.array)

        exp_exact = exact(x)
        exp_exact_dm = exact_dm(x)
        assert jnp.allclose(exp_exact, exp_exact_dm)

        exp_exact_gpsr = exact_gpsr(x)
        exp_exact_gpsr_dm = exact_gpsr_dm(x)
        assert jnp.allclose(exp_exact_gpsr, exp_exact_gpsr_dm)

        assert jnp.allclose(exp_exact, exp_exact_gpsr, atol=GPSR_ATOL)
        assert jnp.allclose(exp_exact_dm, exp_exact_gpsr_dm, atol=GPSR_ATOL)

        exp_shots = shots(x)
        exp_shots_dm = shots_dm(x)

        assert jnp.allclose(exp_exact, exp_shots, atol=SHOTS_ATOL)
        assert jnp.allclose(exp_exact, exp_shots_dm, atol=SHOTS_ATOL)

        d_exact = jax.grad(lambda x: exact(x).sum())
        d_gpsr = jax.grad(lambda x: exact_gpsr(x).sum())
        d_shots = jax.grad(lambda x: shots(x).sum())

        grad_backprop = d_exact(x)
        grad_gpsr = d_gpsr(x)
        assert jnp.allclose(grad_backprop, grad_gpsr, atol=GPSR_ATOL)

        # Check bwd method
        grad_gpsr_bwd = analytical_gpsr_bwd(
            state, list(iter(circuit)), observables, values_to_dict(x)
        )
        assert jnp.allclose(grad_gpsr_bwd, grad_gpsr, atol=1e-3)

        grad_shots = d_shots(x)
        assert jnp.allclose(grad_backprop, grad_shots, atol=SHOTS_ATOL)

        # Check bwd method
        grad_shots_bwd = finite_shots_gpsr_backward(
            state, list(iter(circuit)), observables, values_to_dict(x), n_shots=N_SHOTS
        )
        assert jnp.allclose(grad_shots_bwd, grad_shots, atol=1e-1)

        @self.variant
        def dd_exact(x):
            return jax.hessian(lambda x: exact(x).sum())(x)

        @self.variant
        def dd_gpsr(x):
            return jax.hessian(lambda x: exact_gpsr(x).sum())(x)

        assert jnp.allclose(dd_exact(x), dd_gpsr(x), atol=GPSR_ATOL)

    @chex.variants(with_jit=True, without_jit=True)
    def test_sparse_diff(self) -> None:
        ops = [RX("theta", 0)]
        circuit = QuantumCircuit(2, ops)
        observables = [Observable([Z(0)]), Observable([Z(1)])]
        state = random_state(N_QUBITS)
        x = jnp.pi * 0.5

        @self.variant
        def exact(x):
            values = {"theta": x}
            return expectation(state, circuit, observables, values, diff_mode="ad")

        @self.variant
        def exact_gpsr(x):
            values = {"theta": x}
            return expectation(state, circuit, observables, values, diff_mode="gpsr")

        exp_exact = exact(x)
        exp_gpsr = exact_gpsr(x)

        circuit = to_sparse(circuit)
        observables = [to_sparse(obs) for obs in observables]
        state = random_state(N_QUBITS, sparse=True)

        @self.variant
        def exact_sparse(x):
            values = {"theta": x}
            return expectation(state, circuit, observables, values, diff_mode="ad")

        @self.variant
        def exact_gpsr_sparse(x):
            values = {"theta": x}
            return expectation(state, circuit, observables, values, diff_mode="gpsr")

        exp_exact_sparse = exact_sparse(x)
        exp_gpsr_sparse = exact_gpsr_sparse(x)

        verify_arrays(exp_exact, exp_exact_sparse.todense())
        verify_arrays(exp_gpsr, exp_gpsr_sparse.todense())

        # test gradients raise errors for sparse
        with self.assertRaises(AssertionError):
            jax.experimental.sparse.grad(lambda x: exact_gpsr_sparse(x).sum())(x)



from __future__ import annotations

from functools import singledispatch
from typing import Any

import jax.numpy as jnp
from jax import Array
from jax.experimental.sparse import BCOO

from horqrux.apply import apply_gates, apply_operator
from horqrux.composite import Observable, OpSequence
from horqrux.utils.operator_utils import (
    DensityMatrix,
    OperationType,
    State,
    inner,
    num_qubits,
)
from horqrux.utils.sparse_utils import real_sp, stack_sp
from horqrux.utils.values_utils import _values_processing


@singledispatch
def _ad_expectation_single_observable(
    state: Any,
    observable: Observable,
    values: dict[str, float],
) -> Any:
    raise NotImplementedError("_ad_expectation_single_observable is not implemented")


@_ad_expectation_single_observable.register
def _(
    state: Array,
    observable: Observable,
    values: dict[str, float],
) -> Array:
    projected_state = observable.forward(
        state,
        values,
    )
    return inner(state, projected_state).real


@_ad_expectation_single_observable.register
def _(
    state: BCOO,
    observable: Observable,
    values: dict[str, float],
) -> Array:
    projected_state = observable.forward(
        state,
        values,
    )
    return real_sp(inner(state, projected_state))


@_ad_expectation_single_observable.register
def _(
    state: DensityMatrix,
    observable: Observable,
    values: dict[str, float],
) -> Array:
    n_qubits = num_qubits(state)
    mat_obs = observable.tensor(values)
    d = 2**n_qubits
    prod = apply_operator(state.array, mat_obs, observable.qubit_support, (None,)).reshape((d, d))
    return jnp.trace(prod, axis1=-2, axis2=-1).real


def ad_expectation(
    state: State,
    circuit: OpSequence,
    observables: list[Observable],
    values: dict[str, float] | dict[str, dict[str, float]],
) -> Array:
    """Run 'state' through a sequence of 'gates' given parameters 'values'
       and compute the expectation given an observable.

    Args:
        state (State): Input state vector or density matrix.
        circuit (OpSequence): Sequence of gates.
        observables (list[Observable]): List of observables.
        values (dict[str, float] | dict[str, dict[str, float]]): Parameter values.

    Returns:
        Array: Expectation values.
    """
    values_circuit, values_observables = _values_processing(values)
    outputs = list(
        map(
            lambda observable: _ad_expectation_single_observable(
                apply_gates(
                    state,
                    list(iter(circuit)),  # type: ignore[type-var]
                    values_circuit,
                    OperationType.UNITARY,
                ),
                observable,
                values_observables,
            ),
            observables,
        )
    )
    return stack_sp(outputs)



from __future__ import annotations

import jax.numpy as jnp
from jax.experimental.sparse import sparsify

real_sp = sparsify(jnp.real)
stack_sp = sparsify(jnp.stack)
kron_sp = sparsify(jnp.kron)
eigvalsh_sp = sparsify(jnp.linalg.eigvalsh)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Union

from jax import Array
from jax.tree_util import register_pytree_node_class

from horqrux.utils.operator_utils import StrEnum

from .utils_noise import (
    AmplitudeDamping,
    BitFlip,
    Depolarizing,
    GeneralizedAmplitudeDamping,
    PauliChannel,
    PhaseDamping,
    PhaseFlip,
)


class DigitalNoiseType(StrEnum):
    BITFLIP = "BitFlip"
    PHASEFLIP = "PhaseFlip"
    DEPOLARIZING = "Depolarizing"
    PAULI_CHANNEL = "PauliChannel"
    AMPLITUDE_DAMPING = "AmplitudeDamping"
    PHASE_DAMPING = "PhaseDamping"
    GENERALIZED_AMPLITUDE_DAMPING = "GeneralizedAmplitudeDamping"


PROTOCOL_TO_KRAUS_FN: dict[str, Callable] = {
    "BitFlip": BitFlip,
    "PhaseFlip": PhaseFlip,
    "Depolarizing": Depolarizing,
    "PauliChannel": PauliChannel,
    "AmplitudeDamping": AmplitudeDamping,
    "PhaseDamping": PhaseDamping,
    "GeneralizedAmplitudeDamping": GeneralizedAmplitudeDamping,
}


@register_pytree_node_class
@dataclass
class DigitalNoiseInstance:
    type: DigitalNoiseType
    error_probability: tuple[float, ...] | float

    def tree_flatten(
        self,
    ) -> tuple[tuple, tuple[DigitalNoiseType, tuple[float, ...] | float]]:
        children = ()
        aux_data = (self.type, self.error_probability)
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)

    @property
    def kraus(self) -> tuple[Array, ...]:
        kraus_fn: Callable[..., tuple[Array, ...]] = PROTOCOL_TO_KRAUS_FN[self.type]
        return kraus_fn(error_probability=self.error_probability)

    def __repr__(self) -> str:
        return self.type + f"(p={self.error_probability})"


NoiseProtocol = Union[tuple[DigitalNoiseInstance, ...], None]



from __future__ import annotations


def _values_processing(values: dict | None = None) -> tuple[dict[str, float], dict[str, float]]:
    """Process the parameter values."""
    values = values or dict()
    values_observables = values
    values_circuit = values
    val_keys = values.keys()
    if "circuit" in val_keys and "observables" in val_keys:
        values_observables = values["observables"]
        values_circuit = values["circuit"]
    return values_circuit, values_observables



from __future__ import annotations

from typing import Callable

import jax.numpy as jnp
import numpy as np
import pytest
from jax import Array

from horqrux.apply import apply_gates, apply_operator
from horqrux.primitives.parametric import PHASE, RX, RY, RZ
from horqrux.primitives.primitive import NOT, SWAP, H, I, S, T, X, Y, Z
from horqrux.utils.conversion import to_sparse
from horqrux.utils.operator_utils import OperationType, density_mat, product_state, random_state
from tests.utils import verify_arrays

MAX_QUBITS = 7
PARAMETRIC_GATES = (RX, RY, RZ, PHASE)
PRIMITIVE_GATES = (NOT, H, X, Y, Z, I, S, T)


@pytest.mark.parametrize("gate_fn", PRIMITIVE_GATES)
@pytest.mark.parametrize("sparse", [False, True])
def test_primitive(gate_fn: Callable, sparse: bool) -> None:
    target = np.random.randint(0, MAX_QUBITS)
    gate = gate_fn(target, sparse=sparse)
    orig_state = random_state(MAX_QUBITS, sparse)
    assert len(orig_state) == 2
    state = apply_gates(orig_state, gate)
    assert verify_arrays(
        apply_operator(state, gate.dagger(), gate.target[0], gate.control[0]), orig_state
    )

    # test density matrix is similar to pure state
    if not sparse:
        dm = apply_operator(
            density_mat(orig_state),
            gate._unitary(),
            gate.target[0],
            gate.control[0],
        )
        assert verify_arrays(dm.array, density_mat(state).array)


@pytest.mark.parametrize("gate_fn", PRIMITIVE_GATES)
@pytest.mark.parametrize("sparse", [False, True])
def test_controlled_primitive(gate_fn: Callable, sparse: bool) -> None:
    target = np.random.randint(0, MAX_QUBITS)
    control = np.random.randint(0, MAX_QUBITS)
    while control == target:
        control = np.random.randint(1, MAX_QUBITS)
    gate = gate_fn(target, control, sparse=sparse)
    orig_state = random_state(MAX_QUBITS, sparse)
    state = apply_gates(orig_state, gate)
    assert verify_arrays(
        apply_operator(state, gate.dagger(), gate.target[0], gate.control[0]), orig_state
    )

    # test density matrix is similar to pure state
    if not sparse:
        dm = apply_operator(
            density_mat(orig_state),
            gate._unitary(),
            gate.target[0],
            gate.control[0],
        )
        assert verify_arrays(dm.array, density_mat(state).array)


@pytest.mark.parametrize("gate_fn", PARAMETRIC_GATES)
@pytest.mark.parametrize("sparse", [False, True])
def test_parametric(gate_fn: Callable, sparse: bool) -> None:
    target = np.random.randint(0, MAX_QUBITS)
    gate = gate_fn("theta", target, sparse=sparse)
    values = {"theta": np.random.uniform(0.1, 2 * np.pi)}
    orig_state = random_state(MAX_QUBITS, sparse)
    state = apply_gates(orig_state, gate, values)
    assert verify_arrays(
        apply_operator(state, gate.dagger(values), gate.target[0], gate.control[0]), orig_state
    )

    # test density matrix is similar to pure state
    if not sparse:
        dm = apply_operator(
            density_mat(orig_state),
            gate._unitary(values),
            gate.target[0],
            gate.control[0],
        )
        assert verify_arrays(dm.array, density_mat(state).array)


@pytest.mark.parametrize("gate_fn", PARAMETRIC_GATES)
@pytest.mark.parametrize("sparse", [False, True])
def test_controlled_parametric(gate_fn: Callable, sparse: bool) -> None:
    target = np.random.randint(0, MAX_QUBITS)
    control = np.random.randint(0, MAX_QUBITS)
    while control == target:
        control = np.random.randint(1, MAX_QUBITS)
    gate = gate_fn("theta", target, control, sparse=sparse)
    values = {"theta": np.random.uniform(0.1, 2 * np.pi)}
    orig_state = random_state(MAX_QUBITS, sparse=sparse)
    state = apply_gates(orig_state, gate, values)
    assert verify_arrays(
        apply_operator(state, gate.dagger(values), gate.target[0], gate.control[0]), orig_state
    )

    # test density matrix is similar to pure state
    if not sparse:
        dm = apply_operator(
            density_mat(orig_state),
            gate._unitary(values),
            gate.target[0],
            gate.control[0],
        )
        assert verify_arrays(dm.array, density_mat(state).array)


@pytest.mark.parametrize(
    ["bitstring", "expected_state"],
    [
        ("00", 1 / jnp.sqrt(2) * jnp.array([1, 0, 0, 1])),
        ("01", 1 / jnp.sqrt(2) * jnp.array([0, 1, 1, 0])),
        ("11", 1 / jnp.sqrt(2) * jnp.array([0, 1, -1, 0])),
        ("10", 1 / jnp.sqrt(2) * jnp.array([1, 0, 0, -1])),
    ],
)
@pytest.mark.parametrize("sparse", [False, True])
def test_bell_states(bitstring: str, expected_state: Array, sparse: bool):
    state = product_state(bitstring, sparse)
    state = apply_gates(state, H(target=0, sparse=sparse))
    state = apply_gates(state, NOT(target=1, control=0, sparse=sparse))
    if sparse:
        verify_arrays(state.data, expected_state)
    else:
        verify_arrays(state.flatten(), expected_state)


@pytest.mark.parametrize(
    "inputs",
    [
        ("10", "01", SWAP(target=(0, 1))),
        ("00", "00", SWAP(target=(0, 1))),
        ("001", "100", SWAP(target=(0, 2))),
        ("011", "110", SWAP(target=(0, 2), control=1)),
        ("001", "001", SWAP(target=(0, 2), control=1)),
        ("00101", "01100", SWAP(target=(4, 1), control=2)),
        ("1001001", "1000011", SWAP(target=(5, 3), control=(6, 0))),
    ],
)
@pytest.mark.parametrize("sparse", [False, True])
def test_swap_gate(inputs: tuple[str, str, Array], sparse: bool) -> None:
    bitstring, expected_bitstring, op = inputs
    op.sparse = sparse
    state = product_state(bitstring, sparse=sparse)
    out_state = apply_gates(state, op)
    assert verify_arrays(out_state, product_state(expected_bitstring, sparse=sparse))


@pytest.mark.parametrize("sparse", [False, True])
def test_merge_gates(sparse: bool) -> None:
    gates = [
        RX("a", 0, sparse=sparse),
        RZ("b", 1, sparse=sparse),
        RY("c", 0, sparse=sparse),
        NOT(1, 2, sparse=sparse),
        RX("a", 0, 3, sparse=sparse),
        RZ("c", 3, sparse=sparse),
    ]
    values = {
        "a": np.random.uniform(0.1, 2 * np.pi),
        "b": np.random.uniform(0.1, 2 * np.pi),
        "c": np.random.uniform(0.1, 2 * np.pi),
    }
    state_grouped = apply_gates(
        product_state("0000", sparse=sparse),
        gates,
        values,
        OperationType.UNITARY,
        group_gates=True,
        merge_ops=True,
    )
    state = apply_gates(
        product_state("0000", sparse=sparse),
        gates,
        values,
        OperationType.UNITARY,
        group_gates=False,
        merge_ops=False,
    )
    assert verify_arrays(state_grouped, state)


def flip_bit_wrt_control(bitstring: str, control: int, target: int) -> str:
    # Convert bitstring to list for easier manipulation
    bits = list(bitstring)

    # Flip the bit at the specified index
    if bits[control] == "1":
        bits[target] = "0" if bits[target] == "1" else "1"

    # Convert back to string
    return "".join(bits)


@pytest.mark.parametrize(
    "bitstring",
    [
        "00",
        "01",
        "11",
        "10",
    ],
)
@pytest.mark.parametrize("sparse", [False, True])
def test_cnot_product_state(bitstring: str, sparse: bool):
    cnot0 = NOT(target=1, control=0, sparse=sparse)
    state = product_state(bitstring, sparse=sparse)
    state = apply_gates(state, cnot0)
    expected_state = product_state(flip_bit_wrt_control(bitstring, 0, 1), sparse=sparse)
    assert verify_arrays(state, expected_state)

    # reverse control and target
    cnot1 = NOT(target=0, control=1, sparse=sparse)
    state = product_state(bitstring, sparse=sparse)
    state = apply_gates(state, cnot1)
    expected_state = product_state(flip_bit_wrt_control(bitstring, 1, 0), sparse=sparse)
    assert verify_arrays(state, expected_state)


@pytest.mark.parametrize("sparse", [False, True])
def test_cnot_tensor(sparse: bool) -> None:
    cnot0 = NOT(target=1, control=0, sparse=sparse)
    cnot1 = NOT(target=0, control=1, sparse=sparse)

    t0 = jnp.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
    t0kron = jnp.kron(t0, jnp.eye(2))
    t1 = jnp.array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]])
    t1kron = jnp.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
        ]
    )
    if sparse:
        t0 = to_sparse(t0)
        t1 = to_sparse(t1)
        t0kron = to_sparse(t0kron)
        t1kron = to_sparse(t1kron)

    assert verify_arrays(cnot0.tensor(), t0)
    assert verify_arrays(cnot1.tensor(), t1)

    assert verify_arrays(cnot0.tensor(full_support=(0, 1, 2)), t0kron)
    assert verify_arrays(cnot1.tensor(full_support=(0, 1, 2)), t1kron)


@pytest.mark.parametrize("sparse", [False, True])
def test_crx_tensor(sparse: bool) -> None:
    crx0 = RX(0.2, target=1, control=0, sparse=sparse)
    crx1 = RX(0.2, target=0, control=1, sparse=sparse)
    t0 = jnp.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0.9950, -0.0998j], [0, 0, -0.0998j, 0.9950]])
    t1 = jnp.array([[1, 0, 0, 0], [0, 0.9950, 0, -0.0998j], [0, 0, 1, 0], [0, -0.0998j, 0, 0.9950]])
    if sparse:
        t0 = to_sparse(t0)
        t1 = to_sparse(t1)

    assert verify_arrays(
        crx0.tensor(),
        t0,
        atol=1e-3,
    )
    assert verify_arrays(
        crx1.tensor(),
        t1,
        atol=1e-3,
    )



from __future__ import annotations

from dataclasses import dataclass, field
from functools import reduce
from typing import Any, Iterable

import jax.numpy as jnp
from jax import Array
from jax.tree_util import register_pytree_node_class

from horqrux.utils.operator_utils import State


@register_pytree_node_class
@dataclass
class OpSequence:
    operations: list = field(default_factory=list)

    def tree_flatten(self) -> tuple:
        children = (self.operations,)
        aux_data = ()
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)

    @property
    def qubit_support(self) -> tuple:
        return tuple(set().union(*(op.qubit_support for op in self.operations)))

    def __call__(self, state: State | None = None, values: dict[str, Array] = dict()) -> State:
        return reduce(lambda state, gate: gate.__call__(state, values), self.operations, state)

    def tensor(
        self,
        values: dict[str, float] = dict(),
        full_support: tuple[int, ...] | None = None,
    ) -> Array:
        """Obtain the unitary.

        Args:
            values (dict[str, float], optional): Parameter values. Defaults to dict().
            full_support (tuple[int, ...], optional): The qubit support of definition for the unitary.

        Returns:
            Array: Unitary representation.
        """
        if full_support is None:
            full_support = self.qubit_support
        elif not set(self.qubit_support).issubset(set(full_support)):
            raise ValueError(
                "Expanding tensor operation requires a `full_support` argument "
                "larger than or equal to the `qubit_support`."
            )
        return reduce(jnp.matmul, map(lambda x: x.tensor(values, full_support), self.operations))

    def __iter__(self) -> Iterable:
        def flatten(item: Any) -> Iterable:
            # If the item is a OpSequence, iterate through its operations
            if isinstance(item, OpSequence):
                for sub_item in item.operations:
                    yield from flatten(sub_item)
            # If the item is not a OpSequence, yield it directly
            else:
                yield item

        # Iterate through the operations and flatten
        for operation in self.operations:
            yield from flatten(operation)

    @property
    def param_names(self) -> set[str]:
        """List of parameters of the circuit.
            Composed of variational and feature map parameters.

        Returns:
            set[str]: Names of parameters.
        """
        from horqrux.primitives import Parametric

        return set([str(op.param) for op in list(iter(self)) if isinstance(op, Parametric)])  # type: ignore[type-var]

    @property
    def n_vparams(self) -> int:
        """Number of variational parameters.

        Returns:
            int: Number of variational parameters.
        """
        return len(self.param_names)



from __future__ import annotations

import jax.numpy as jnp
from jax import config
from jax.experimental.sparse import BCOO

from ._misc import default_complex_dtype

config.update("jax_enable_x64", True)  # Quantum ML requires higher precision
default_dtype = default_complex_dtype()

_X = jnp.array([[0, 1], [1, 0]], dtype=default_dtype)
_Y = jnp.array([[0, -1j], [1j, 0]], dtype=default_dtype)
_Z = jnp.array([[1, 0], [0, -1]], dtype=default_dtype)
_H = jnp.array([[1, 1], [1, -1]], dtype=default_dtype) * 1 / jnp.sqrt(2)
_S = jnp.array([[1, 0], [0, 1j]], dtype=default_dtype)
_T = jnp.array([[1, 0], [0, jnp.exp(1j * jnp.pi / 4)]], dtype=default_dtype)
_I = jnp.asarray([[1, 0], [0, 1]], dtype=default_dtype)

_SWAP = jnp.asarray([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]], dtype=default_dtype)


_SQSWAP = jnp.asarray(
    [
        [1, 0, 0, 0],
        [0, 0.5 * (1 + 1j), 0.5 * (1 - 1j), 0],
        [0, 0.5 * (1 - 1j), 0.5 * (1 + 1j), 0],
        [0, 0, 0, 1],
    ],
    dtype=default_dtype,
)

_ISWAP = jnp.asarray(
    [[1, 0, 0, 0], [0, 0, 1j, 0], [0, 1j, 0, 0], [0, 0, 0, 1]], dtype=default_dtype
)

_ISQSWAP = jnp.asarray(
    [
        [1, 0, 0, 0],
        [0, 1 / jnp.sqrt(2), 1j / jnp.sqrt(2), 0],
        [0, 1j / jnp.sqrt(2), 1 / jnp.sqrt(2), 0],
        [0, 0, 0, 1],
    ],
    dtype=default_dtype,
)

OPERATIONS_DICT = {
    "X": _X,
    "Y": _Y,
    "Z": _Z,
    "H": _H,
    "S": _S,
    "T": _T,
    "I": _I,
    "SWAP": _SWAP,
}

SPARSE_OPERATIONS_DICT = {k: BCOO.fromdense(mat) for k, mat in OPERATIONS_DICT.items()}



from __future__ import annotations

from functools import partial
from typing import Iterable, Union

import jax
import jax.numpy as jnp
from jax import Array

from horqrux.apply import apply_gates
from horqrux.composite import Observable
from horqrux.differentiation.ad import _ad_expectation_single_observable
from horqrux.differentiation.gpsr.gpsr_utils import (
    alter_gate_sequence,
    create_renamed_operators,
    extract_gate_names,
    initialize_gpsr_ingredients,
    prepare_param_gates_seq,
    spectral_gap_from_gates,
)
from horqrux.primitives import Primitive
from horqrux.utils.operator_utils import State
from horqrux.utils.sparse_utils import stack_sp


def analytical_expectation(
    state: State,
    gates: Union[Primitive, Iterable[Primitive]],
    observables: list[Observable],
    values: dict[str, float],
) -> Array:
    """Run 'state' through a sequence of 'gates' given parameters 'values'
    and compute the expectations analytically.

    Args:
        state (State): Input state or density matrix.
        gates (Union[Primitive, Iterable[Primitive]]): Sequence of gates.
        observables (list[Observable]): List of observables.
        values (dict[str, float]): Parameter values.

    Returns:
        Array: Expectation values.
    """
    outputs = list(
        map(
            lambda observable: _ad_expectation_single_observable(
                apply_gates(state, gates, values),
                observable,
                values,
            ),
            observables,
        )
    )
    return stack_sp(outputs)


jitted_analytical_exp = jax.jit(analytical_expectation)


def analytical_gpsr_bwd(
    state: State,
    gates: Union[Primitive, Iterable[Primitive]],
    observables: list[Observable],
    values: dict[str, float],
) -> Array:
    """Compute the gradient using the GPSR algorithm.
    This function is efficient when used in a
    training loop using the jitted version
    of the `no_shots` function
    (that is the `horqrux.differentiation.gpsr.jitted_analytical_exp`).

    There is also a jvp function `no_shots_fwd`compatible with `jax.grad`
    but this function may be faster when using `jax.jit`.

    Args:
        state (State): Input state or density matrix.
        gates (Union[Primitive, Iterable[Primitive]]): Sequence of gates.
        observables (list[Observable]): List of observables.
        values (dict[str, float]): Parameter values.

    Returns:
        Array: Gradients via the GPSR algorithm.
    """
    val_keys, spectral_gap, shift = initialize_gpsr_ingredients(values)

    def jvp_component(param_name: str) -> Array:
        up_val = values.copy()
        up_val[param_name] = up_val[param_name] + shift
        f_up = jitted_analytical_exp(state, gates, observables, up_val)
        down_val = values.copy()
        down_val[param_name] = down_val[param_name] - shift
        f_down = jitted_analytical_exp(state, gates, observables, down_val)
        grad = (
            spectral_gap[param_name]
            * (f_up - f_down)
            / (4.0 * jnp.sin(spectral_gap[param_name] * shift / 2.0))
        )
        return grad

    jvp_caller = jvp_component
    if not isinstance(gates, Primitive):
        gate_names = extract_gate_names(gates)
        if len(gate_names) > len(val_keys):
            param_to_gates_indices = prepare_param_gates_seq(val_keys, gates)
            # repeated case
            if max(map(len, param_to_gates_indices.values())) > 1:  # type: ignore[arg-type]

                def jvp_component_repeated_param(param_name: str) -> Array:
                    shift_gates = param_to_gates_indices[param_name]

                    def shift_jvp(ind: int) -> Array:
                        spectral_gap = gates[ind].spectral_gap  # type: ignore[index]
                        gates_up = alter_gate_sequence(gates, ind, shift)
                        f_up = jitted_analytical_exp(state, gates_up, observables, values)
                        gates_down = alter_gate_sequence(gates, ind, -shift)
                        f_down = jitted_analytical_exp(state, gates_down, observables, values)
                        return (
                            spectral_gap
                            * (f_up - f_down)
                            / (4.0 * jnp.sin(spectral_gap * shift / 2.0))
                        )

                    return sum(shift_jvp(shift_ind) for shift_ind in shift_gates)

                jvp_caller = jvp_component_repeated_param
            else:
                spectral_gap = spectral_gap_from_gates(param_to_gates_indices, val_keys)

    jvp = jnp.stack([jvp_caller(param) for param in val_keys]).sum(axis=-1)
    return jvp


@partial(jax.custom_jvp, nondiff_argnums=(0, 1, 2))
def analytical_gpsr_fwd(
    state: State,
    gates: Union[Primitive, Iterable[Primitive]],
    observables: list[Observable],
    values: dict[str, float],
) -> Array:
    """Run 'state' through a sequence of 'gates' given parameters 'values'
    and compute the expectations analytically.

    This is compatible with `jax.grad`as we use the `jax.custom_jvp` for
    setting a custom differentiation rule to this function.

    Note though we de not recommend using `jax.jit` with it as compilation
    time is very long.

    Args:
        state (State): Input state or density matrix.
        gates (Union[Primitive, Iterable[Primitive]]): Sequence of gates.
        observables (list[Observable]): List of observables.
        values (dict[str, float]): Parameter values.

    Returns:
        Array: Expectation values.
    """
    return analytical_expectation(state, gates, observables, values)


@analytical_gpsr_fwd.defjvp
def analytical_gpsr_jvp(
    state: Array,
    gates: Union[Primitive, Iterable[Primitive]],
    observable: list[Observable],
    primals: tuple[dict[str, float]],
    tangents: tuple[dict[str, float]],
) -> Array:
    """Jvp version for `analytical_gpsr_fwd`.

    Args:
        state (Array): Input state or density matrix.
        gates (Union[Primitive, Iterable[Primitive]]): sequence of gates.
        observable (list[Observable]): List of observables.
        primals (tuple[dict[str, Array]]): Values we are differentiating over.
            Jax-specific syntax argument.
        tangents (tuple[dict[str, Array]]): Bases for differentiation.
            Jax-specific syntax argument.

    Returns:
        tuple[Array, Array]: Forward eveluation and gradient
    """
    values = primals[0]
    tangent_dict = tangents[0]
    fwd = analytical_gpsr_fwd(state, gates, observable, values)

    legit_val_keys, spectral_gap, shift = initialize_gpsr_ingredients(values)
    values_map = None

    if not isinstance(gates, Primitive):
        gate_names = extract_gate_names(gates)
        if len(gate_names) > len(legit_val_keys):
            param_to_gates_indices = prepare_param_gates_seq(legit_val_keys, gates)
            # repeated case
            if max(map(len, param_to_gates_indices.values())) > 1:  # type: ignore[arg-type]
                # use temporary gates and values to make use of scan
                gates = create_renamed_operators(gates, legit_val_keys)  # type: ignore[index]
                values_map = {
                    gates[ind].param: pname  # type: ignore[index]
                    for pname in legit_val_keys
                    for ind in param_to_gates_indices[pname]
                }
                values = {
                    temp_name: values[values_map[temp_name]] for temp_name in values_map.keys()
                }
                spectral_gap = {
                    temp_name: gates[ind].spectral_gap  # type: ignore[index]
                    for temp_name in values_map.keys()
                    for ind in param_to_gates_indices[values_map[temp_name]]
                }
            else:
                spectral_gap = spectral_gap_from_gates(param_to_gates_indices, legit_val_keys)

    values_array = stack_sp(list(values.values()))
    if values_array.shape[-1] == 1:
        values_array = values_array.squeeze(-1)
    vals_to_dict = values.keys()

    def values_to_dict(x: Array) -> dict[str, Array]:
        return dict(zip(vals_to_dict, x))

    def jvp_caller(carry_grad: Array, pytree_ind_param: dict[str, Array]) -> tuple[Array, Array]:
        shift_vector = pytree_ind_param["shift"]
        spectral_gap = pytree_ind_param["spectral_gap"]
        up_val = values_to_dict(values_array + shift_vector)
        f_up = analytical_gpsr_fwd(state, gates, observable, up_val)
        down_val = values_to_dict(values_array - shift_vector)
        f_down = analytical_gpsr_fwd(state, gates, observable, down_val)
        grad = spectral_gap * (f_up - f_down) / (4.0 * jnp.sin(spectral_gap * shift / 2.0))
        return carry_grad + grad, grad

    spectral_gap_array = stack_sp(list(spectral_gap.values()))
    tangent_array = stack_sp(list(tangent_dict.values())).reshape((-1, 1))
    pytree_scan = {
        "shift": shift * jnp.eye(len(vals_to_dict), dtype=values_array.dtype),
        "spectral_gap": spectral_gap_array,
    }

    _, grads = jax.lax.scan(jvp_caller, jnp.zeros(fwd.shape), pytree_scan)
    if values_map:
        # need to remap to original parameter names
        grad_dict = dict(zip(values.keys(), grads))
        grads_legit_dict: dict = {name: list() for name in legit_val_keys}
        for temp_name in grad_dict.keys():
            grads_legit_dict[values_map[temp_name]].append(grad_dict[temp_name])
        grads = tuple(stack_sp(grads_legit_dict[name]).sum(axis=0) for name in legit_val_keys)
    jvp = (stack_sp(grads) * tangent_array).sum(axis=0)
    return fwd, jvp.reshape(fwd.shape)



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from enum import Enum
from functools import reduce, singledispatch
from math import log
from typing import Any, Iterable, Union

import jax
import jax.numpy as jnp
import numpy as np
from jax import Array
from jax.experimental.sparse import BCOO, sparsify
from jax.tree_util import register_pytree_node_class
from jax.typing import ArrayLike
from numpy import log2

from horqrux._misc import default_complex_dtype
from horqrux.matrices import _I

from .sparse_utils import kron_sp

default_dtype = default_complex_dtype()

QubitSupport = tuple[Any, ...]
ControlQubits = tuple[Union[None, tuple[int, ...]], ...]
TargetQubits = tuple[tuple[int, ...], ...]
ATOL = 1e-014


@register_pytree_node_class
@dataclass
class DensityMatrix:
    """Dataclass to identify density matrices from states."""

    array: Array

    def tree_flatten(self) -> tuple[tuple, tuple[Array]]:
        children = ()
        aux_data = (self.array,)
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)


State = Union[ArrayLike, DensityMatrix]


def density_mat(state: ArrayLike) -> DensityMatrix:
    """Convert state to density matrix

    Args:
        state (ArrayLike): Input state.

    Returns:
        DensityMatrix: Density matrix representation.
    """
    # Expand dimensions to enable broadcasting
    if isinstance(state, DensityMatrix):
        return state
    ket = jnp.expand_dims(state, axis=tuple(range(state.ndim, 2 * state.ndim)))
    bra = jnp.conj(jnp.expand_dims(state, axis=tuple(range(state.ndim))))
    return DensityMatrix(ket * bra)


def permute_basis(operator: Array | BCOO, qubit_support: tuple, inv: bool = False) -> Array | BCOO:
    """Takes an operator tensor and permutes the rows and
    columns according to the order of the qubit support.

    Args:
        operator (Array | BCOO): Operator to permute over.
        qubit_support (tuple): Qubit support.
        inv (bool): Applies the inverse permutation instead.

    Returns:
        Array | BCOO: Permuted operator.
    """
    ordered_support = np.argsort(qubit_support)
    ranked_support = np.argsort(ordered_support)
    n_qubits = len(qubit_support)
    if all(a == b for a, b in zip(ranked_support, tuple(range(n_qubits)))):
        return operator

    perm = tuple(ranked_support) + tuple(ranked_support + n_qubits)
    if inv:
        perm = np.argsort(perm)
    transpose_fn = (
        jax.experimental.sparse.bcoo_transpose if isinstance(operator, BCOO) else jax.lax.transpose
    )
    return transpose_fn(operator, permutation=perm)


class StrEnum(str, Enum):
    def __str__(self) -> str:
        """Used when dumping enum fields in a schema."""
        ret: str = self.value
        return ret

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda c: c.value, cls))  # type: ignore


class OperationType(StrEnum):
    UNITARY = "_unitary"
    DAGGER = "dagger"
    JACOBIAN = "jacobian"


class DiffMode(StrEnum):
    """Differentiation mode."""

    AD = "ad"
    """Automatic Differentiation -  Using the autograd engine of JAX."""
    ADJOINT = "adjoint"
    """Adjoint Differentiation   - An implementation of "Efficient calculation of gradients
                                   in classical simulations of variational quantum algorithms",
                                   Jones & Gacon, 2020."""
    GPSR = "gpsr"
    """Generalized parameter shift rule."""


@sparsify
def _dagger(operator: Array) -> Array:
    # If the operator is a tensor with repeated 2D axes
    if operator.ndim > 2:
        # Conjugate and swap the last two axes
        conjugated = operator.conj()

        # Create the transpose axes: swap pairs of indices
        half = operator.ndim // 2
        axes = tuple(range(half, operator.ndim)) + tuple(range(half))
        return jnp.transpose(conjugated, axes)
    else:
        # For standard matrices, use conjugate transpose
        return jnp.conjugate(operator.T)


@singledispatch
def _unitary(generator: Any, theta: float) -> Any:
    raise NotImplementedError("_unitary is not implemented")


@_unitary.register
def _(generator: Array, theta: float) -> Array:
    return (
        jnp.cos(theta / 2) * jnp.eye(2, dtype=default_dtype) - 1j * jnp.sin(theta / 2) * generator
    )


@_unitary.register
def _(generator: BCOO, theta: float) -> BCOO:
    return (
        jnp.cos(theta / 2) * jax.experimental.sparse.eye(2, dtype=default_dtype)
        - 1j * jnp.sin(theta / 2) * generator
    )


@singledispatch
def _jacobian(generator: Any, theta: float) -> Any:
    raise NotImplementedError("_jacobian is not implemented")


@_jacobian.register
def _(generator: Array, theta: float) -> Array:
    return (
        -1
        / 2
        * (jnp.sin(theta / 2) * jnp.eye(2, dtype=default_dtype) + 1j * jnp.cos(theta / 2))
        * generator
    )


@_jacobian.register
def _(generator: BCOO, theta: float) -> BCOO:
    return (
        -1
        / 2
        * (
            jnp.sin(theta / 2) * jax.experimental.sparse.eye(2, dtype=default_dtype)
            + 1j * jnp.cos(theta / 2)
        )
        * generator
    )


@singledispatch
def _controlled(operator: Any, n_control: int) -> Any:
    raise NotImplementedError("_controlled is not implemented")


@_controlled.register
def _(operator: Array, n_control: int) -> Array:
    """
    Create a controlled quantum operator with specified number of control qubits.

    Args:
        operator (jnp.ndarray): The base quantum operator to be controlled.
        n_control (int): Number of control qubits.

    Returns:
        jnp.ndarray: The controlled quantum operator matrix
    """
    n_qubits = int(log2(operator.shape[0]))
    control = jnp.eye(2 ** (n_control + n_qubits), dtype=default_dtype)
    control = control.at[-(2**n_qubits) :, -(2**n_qubits) :].set(operator)
    return control


@_controlled.register
def _(operator: BCOO, n_control: int) -> BCOO:
    """
    Create a controlled quantum operator with specified number of control qubits.

    Args:
        operator (jnp.ndarray): The base quantum operator to be controlled.
        n_control (int): Number of control qubits.

    Returns:
        jnp.ndarray: The controlled quantum operator matrix
    """
    n_qubits = int(log2(operator.shape[0]))
    shape_dim = 2 ** (n_control + n_qubits)
    control_elements = shape_dim - 2**n_qubits
    control_id = jax.experimental.sparse.eye(control_elements, dtype=default_dtype)
    data_control = jnp.concatenate((control_id.data, operator.data))
    indices_control = jnp.concatenate((control_id.indices, operator.indices + control_elements))
    control = jax.experimental.sparse.BCOO(
        (data_control, indices_control), shape=(shape_dim, shape_dim)
    )
    return control


@singledispatch
def controlled(
    operator: Any,
    target_qubits: TargetQubits,
    control_qubits: ControlQubits,
) -> Any:
    raise NotImplementedError(
        f"Controlled is not implemented for this operator type: {type(operator)}."
    )


def prepare_controlled(
    operator: ArrayLike, target_qubits: TargetQubits, control_qubits: ControlQubits
) -> tuple[int, ArrayLike]:
    controls: tuple = tuple()
    targets: tuple = tuple()
    if isinstance(control_qubits[0], tuple):
        controls = control_qubits[0]
    if isinstance(target_qubits[0], tuple):
        targets = target_qubits[0]
    n_qop = int(log(operator.shape[0], 2))
    n_targets = len(targets)
    if n_qop != n_targets:
        raise ValueError("`target_qubits` length should match the shape of operator.")
    # Determine the total number of qubits and order of controls
    ntotal_qubits = len(controls) + n_targets
    qubit_support = sorted(controls + targets)
    control_ind_support = tuple(i for i, q in enumerate(qubit_support) if q in controls)

    # Create the full Hilbert space dimension
    full_dim = 2**ntotal_qubits

    # Compute the control mask using bit manipulation
    control_mask = jnp.sum(
        jnp.array(
            [1 << (ntotal_qubits - control_qubit - 1) for control_qubit in control_ind_support]
        )
    )

    # Create indices for the controlled subspace
    indices = jnp.arange(full_dim)
    controlled_indices = indices[(indices & control_mask) == control_mask]
    return full_dim, controlled_indices


@controlled.register
def _(
    operator: Array,
    target_qubits: TargetQubits,
    control_qubits: ControlQubits,
) -> Array:
    """
    Create a controlled quantum operator with specified control and target qubit indices.

    Args:
        operator (Array): The base quantum operator to be controlled.
            Note the operator is defined only on `target_qubits`.
        control_qubits (int or tuple of ints): Index or indices of control qubits
        target_qubits (int or tuple of ints): Index or indices of target qubits

    Returns:
        Array: The controlled quantum operator matrix
    """
    full_dim, controlled_indices = prepare_controlled(operator, target_qubits, control_qubits)
    # Initialize the controlled operator as an identity matrix
    # and set the controlled subspace to the operator
    controlled_op = jnp.eye(full_dim, dtype=operator.dtype)
    controlled_op = controlled_op.at[jnp.ix_(controlled_indices, controlled_indices)].set(operator)

    return controlled_op


@controlled.register
def _(
    operator: BCOO,
    target_qubits: TargetQubits,
    control_qubits: ControlQubits,
) -> BCOO:
    """
    Create a controlled quantum operator with specified control and target qubit indices.

    Args:
        operator (Array): The base quantum operator to be controlled.
            Note the operator is defined only on `target_qubits`.
        control_qubits (int or tuple of ints): Index or indices of control qubits
        target_qubits (int or tuple of ints): Index or indices of target qubits

    Returns:
        Array: The controlled quantum operator matrix
    """
    full_dim, controlled_indices = prepare_controlled(operator, target_qubits, control_qubits)

    # Initialize the controlled operator as an identity matrix
    # and set the controlled subspace to the operator
    # TODO: optimize this
    controlled_op = jnp.eye(full_dim, dtype=operator.dtype)
    controlled_op = controlled_op.at[jnp.ix_(controlled_indices, controlled_indices)].set(
        operator.todense()
    )
    return BCOO.fromdense(controlled_op)


def expand_operator(
    operator: Array | BCOO, qubit_support: tuple[int, ...], full_support: tuple[int, ...]
) -> Array | BCOO:
    """
    Expands an operator acting on a given qubit_support to act on a larger full_support
    by explicitly filling in identity matrices on all remaining qubits.

    Args:
        operator (Array | BCOO): Operator to expand
        qubit_support (tuple[int, ...]): Qubit support the operator is initially defined over.
        full_support (tuple[int, ...]): Qubit support the operator will be defined over.

    Raises:
        ValueError: When `full_support` larger than or equal to the `qubit_support`

    Returns:
        Array | BCOO: Expanded operator.
    """
    full_support = tuple(sorted(full_support))
    qubit_support = tuple(sorted(qubit_support))
    n_full = 2 ** len(full_support)
    if not set(qubit_support).issubset(set(full_support)):
        raise ValueError(
            "Expanding tensor operation requires a `full_support` argument "
            "larger than or equal to the `qubit_support`."
        )

    kron_qubits = tuple(sorted(set(full_support) - set(qubit_support)))
    if isinstance(operator, BCOO):
        kron_operator = reduce(
            kron_sp,
            [operator] + [jax.experimental.sparse.eye(2, dtype=default_dtype)] * len(kron_qubits),
        )
    else:
        kron_operator = reduce(jnp.kron, [operator] + [_I] * len(kron_qubits))

    kron_operator = hilbert_reshape(kron_operator)
    kron_operator = permute_basis(kron_operator, qubit_support + kron_qubits, True)
    kron_operator = kron_operator.reshape((n_full, n_full))
    return kron_operator


def product_state(bitstring: str, sparse: bool = False) -> Array:
    """Generates a state of shape [2 for _ in range(len(bitstring))].

    Args:
        bitstring: The target bitstring.

    Returns:
        A state corresponding to 'bitstring'.
    """
    n_qubits = len(bitstring)
    shape = tuple(2 for _ in range(n_qubits))
    bitstr_ind = tuple(map(int, bitstring))
    if sparse:
        data = jnp.array([1.0], dtype=default_dtype)
        indices = jnp.array([bitstr_ind])
        space = BCOO((data, indices), shape=shape)
    else:
        space = jnp.zeros(shape, dtype=default_dtype)
        space = space.at[bitstr_ind].set(1.0)
    return space


def zero_state(n_qubits: int, sparse: bool = False) -> Array:
    return product_state("0" * n_qubits, sparse)


def none_like(x: Iterable) -> tuple[None, ...]:
    """Generates a tuple of Nones with equal length to x. Useful for gates with multiple targets but no control.

    Args:
        x (Iterable): Iterable to be mimicked.

    Returns:
        tuple[None, ...]: tuple of Nones of length x.
    """
    return tuple(map(lambda _: None, x))


def hilbert_reshape(operator: ArrayLike) -> Array:
    """Reshapes 'operator' of shape [M, M] to array of shape [2, 2, ...].
       Useful for working with controlled and multi-qubit gates.

    Args:
        operator (Array): Array to be reshaped.

    Returns:
        Array: Array with shape [2, 2, ...]
    """
    n_axes = int(np.log2(operator.size))
    return operator.reshape(tuple(2 for _ in np.arange(n_axes)))


def equivalent_state(s0: Array, s1: Array) -> bool:
    """Utility to easily compare two states.

    Args:
        s0: State array to be checked.
        s1:  State array to compare to.

    Returns:
        bool: Boolean indicating whether the states are equivalent.
    """
    return jnp.allclose(overlap(s0, s1), 1.0, atol=ATOL)  # type: ignore[no-any-return]


@sparsify
def inner(state: Array, projection: Array) -> Array:
    return jnp.dot(jnp.conj(state.flatten()), projection.flatten())


def overlap(state: Array, projection: Array) -> Array:
    return jnp.power(inner(state, projection), 2).real


def uniform_state(n_qubits: int, sparse: bool = False) -> ArrayLike:
    state = jnp.ones(2**n_qubits, dtype=default_dtype)
    state = state / jnp.sqrt(jnp.array(2**n_qubits, dtype=default_dtype))
    if sparse:
        state = BCOO.fromdense(state)
    return state.reshape([2] * n_qubits)


def is_controlled(qubit_support: Union[tuple[Union[int, None], ...], int, None]) -> bool:
    if isinstance(qubit_support, int):
        return True
    elif isinstance(qubit_support, tuple):
        return any(is_controlled(q) for q in qubit_support)
    return False


def random_state(n_qubits: int, sparse: bool = False) -> ArrayLike:
    def _normalize(wf: Array) -> Array:
        return wf / jnp.sqrt((jnp.sum(jnp.abs(wf) ** 2)))

    key = jax.random.PRNGKey(n_qubits)
    N = 2**n_qubits
    x = -jnp.log(jax.random.uniform(key, shape=(N,)))
    sumx = jnp.sum(x)
    phases = jax.random.uniform(key, shape=(N,)) * 2.0 * jnp.pi
    state = _normalize(
        (jnp.sqrt(x / sumx) * jnp.exp(1j * phases)).reshape(tuple(2 for _ in range(n_qubits)))
    )
    if sparse:
        state = BCOO.fromdense(state)
    return state


def is_normalized(state: Array) -> bool:
    return equivalent_state(state, state)


def sample_from_probs(probs: Array, n_qubits: int, n_shots: int) -> Counter:
    key = jax.random.PRNGKey(0)

    # JAX handles pseudo random number generation by tracking an explicit state via a random key
    # For more details, see https://jax.readthedocs.io/en/latest/random-numbers.html
    samples = jax.vmap(
        lambda subkey: jax.random.choice(key=subkey, a=jnp.arange(0, 2**n_qubits), p=probs)
    )(jax.random.split(key, n_shots))

    return Counter(
        {
            format(k, f"0{n_qubits}b"): count.item()
            for k, count in enumerate(jnp.bincount(samples))
            if count > 0
        }
    )


@singledispatch
def probabilities(state: Any) -> Array:
    """Extract probabilities from state or density matrix.

    Args:
        state (Array): Input array.

    Raises:
        NotImplementedError: If not implemented for given types.

    Returns:
        Array: Vector of probabilities.
    """
    raise NotImplementedError(f"Probabilities is not implemented for the input type {type(state)}.")


@probabilities.register
def _(state: Array) -> Array:
    return jnp.abs(jnp.float_power(state, 2.0)).ravel()


@probabilities.register
def _(state: DensityMatrix) -> Array:
    return jnp.diagonal(state.array).real


@singledispatch
def num_qubits(state: Any) -> int:
    """Returns the number of qubits of a state.

    Args:
        state (Any): state.

    Returns:
        int: Number of qubits.
    """
    raise NotImplementedError(f"num_qubits is not implemented for the state type {type(state)}.")


@num_qubits.register
def _(state: Array) -> int:
    return len(state.shape)


@num_qubits.register
def _(state: DensityMatrix) -> int:
    return len(state.array.shape) // 2



from __future__ import annotations

from collections import Counter
from typing import Any

import jax
from jax import Array

from horqrux.composite import Observable, OpSequence
from horqrux.differentiation.ad import ad_expectation
from horqrux.differentiation.adjoint import adjoint_expectation as apply_adjoint
from horqrux.differentiation.gpsr import analytical_gpsr_fwd, finite_shots_fwd
from horqrux.shots import finite_shots
from horqrux.utils.operator_utils import (
    DensityMatrix,
    DiffMode,
    State,
    num_qubits,
    probabilities,
    sample_from_probs,
)
from horqrux.utils.sparse_utils import stack_sp


def run(
    circuit: OpSequence,
    state: State,
    values: dict[str, float] = dict(),
) -> State:
    return circuit(state, values)


def sample(
    state: State,
    circuit: OpSequence,
    values: dict[str, float] = dict(),
    n_shots: int = 1000,
) -> Counter:
    """Sample from a quantum program.

    Args:
        state (State): Input state vector or density matrix.
        circuit (OpSequence): Sequence of gates.
        values (dict[str, float], optional): _description_. Defaults to dict().
        n_shots (int, optional): Parameter values.. Defaults to 1000.

    Raises:
        ValueError: If n_shots < 1.

    Returns:
        Counter: Bitstrings and frequencies.
    """
    if n_shots < 1:
        raise ValueError("You can only sample with non-negative 'n_shots'.")
    output_circuit = circuit(state, values)
    n_qubits = num_qubits(output_circuit)
    if isinstance(output_circuit, DensityMatrix):
        d = 2**n_qubits
        output_circuit.array = output_circuit.array.reshape((d, d))

    probs = probabilities(output_circuit)
    return sample_from_probs(probs, n_qubits, n_shots)


def adjoint_expectation(
    state: State,
    circuit: OpSequence,
    observables: list[Observable],
    values: dict | dict[str, float] | dict[str, dict[str, float]] = dict(),
) -> Array:
    """Apply a sequence of adjoint operators to an input state given parameters 'values'
       and compute the expectation given an observable.

    Args:
        state (State): Input state vector or density matrix.
        circuit (OpSequence): Sequence of gates.
        observables (list[Observable]): List of observables.
        values (dict, dict[str, float] | dict[str, dict[str, float]], optional): Parameter values.

    Returns:
        Array: Expectation values.
    """
    outputs = list(
        map(
            lambda observable: apply_adjoint(state, circuit, observable, values),
            observables,
        )
    )
    return stack_sp(outputs)


def expectation(
    state: State,
    circuit: OpSequence,
    observables: list[Observable],
    values: dict | dict[str, float] | dict[str, dict[str, float]] = dict(),
    diff_mode: DiffMode = DiffMode.AD,
    n_shots: int = 0,
    key: Any = jax.random.PRNGKey(0),
) -> Array:
    """Run 'state' through a sequence of 'gates' given parameters 'values'
    and compute the expectation given an observable.

    Args:
        state (State): Input state vector or density matrix.
        circuit (OpSequence): Sequence of gates.
        observables (list[Observable]): List of observables.
        values (dict[str, float] | dict[str, dict[str, float]], optional): A dictionary
            containing <'parameter_name': value> pairs
            denoting the current parameter values for each parameter in `circuit`.
            Note it can include also values for the observables, but differentiation will
            not separate gradients.
            To do so, we should provide values as a dict with two keys: `circuit` and `observables`, each a dict.
        diff_mode (DiffMode, optional): Differentiation mode. Defaults to DiffMode.AD.
        n_shots (int): Number of shots. Defaults to 0 for no shots.
        key (Any, optional): Random key. Defaults to jax.random.PRNGKey(0).

    Returns:
        Array: Expectation values.
    """

    if diff_mode == DiffMode.AD:
        if n_shots < 0:
            raise ValueError("The number of shots should be positive.")
        if n_shots == 0:
            return ad_expectation(state, circuit, observables, values)
        else:
            return finite_shots(
                state=state,
                gates=list(iter(circuit)),  # type: ignore[type-var]
                observables=observables,
                values=values,  # type: ignore[arg-type]
                n_shots=n_shots,
                key=key,
            )
    elif diff_mode == DiffMode.ADJOINT:
        if isinstance(state, DensityMatrix):
            raise TypeError("Adjoint does not support density matrices.")

        return adjoint_expectation(state, circuit, observables, values)

    elif diff_mode == DiffMode.GPSR:
        if n_shots < 0:
            raise ValueError("The number of shots should be positive.")
        if n_shots == 0:
            return analytical_gpsr_fwd(
                state=state,
                gates=list(iter(circuit)),  # type: ignore[type-var]
                observables=observables,
                values=values,
            )
        else:
            return finite_shots_fwd(
                state=state,
                gates=list(iter(circuit)),  # type: ignore[type-var]
                observables=observables,
                values=values,
                n_shots=n_shots,
                key=key,
            )
    else:
        raise ValueError(f"Differentiation mode {diff_mode} is not supported.")



from __future__ import annotations

from typing import Callable

import jax.numpy as jnp
import numpy as np
import pytest

from horqrux.api import expectation, run, sample
from horqrux.apply import apply_gates
from horqrux.noise import DigitalNoiseInstance, DigitalNoiseType
from horqrux.primitives.parametric import PHASE, RX, RY, RZ
from horqrux.primitives.primitive import NOT, H, I, S, T, X, Y, Z
from horqrux.utils.operator_utils import density_mat, product_state, random_state

MAX_QUBITS = 7
PARAMETRIC_GATES = (RX, RY, RZ, PHASE)
PRIMITIVE_GATES = (NOT, H, X, Y, Z, I, S, T)

NOISE_single_prob = (
    DigitalNoiseType.BITFLIP,
    DigitalNoiseType.PHASEFLIP,
    DigitalNoiseType.DEPOLARIZING,
    DigitalNoiseType.AMPLITUDE_DAMPING,
    DigitalNoiseType.PHASE_DAMPING,
)
ALL_NOISES = list(DigitalNoiseType)


def noise_instance(noise_type: DigitalNoiseType, zero_proba: bool = False) -> DigitalNoiseInstance:
    if noise_type in NOISE_single_prob:
        errors = 0.1 if not zero_proba else 0.0
    elif noise_type == DigitalNoiseType.PAULI_CHANNEL:
        errors = (0.4, 0.5, 0.1) if not zero_proba else (0.0,) * 3
    else:
        errors = (0.2, 0.8) if not zero_proba else (0.0,) * 2

    return DigitalNoiseInstance(noise_type, error_probability=errors)


@pytest.mark.parametrize("noise_type", NOISE_single_prob)
def test_error_prob(noise_type: DigitalNoiseType):
    with pytest.raises(ValueError):
        noise = DigitalNoiseInstance(noise_type, error_probability=-0.5).kraus
    with pytest.raises(ValueError):
        noise = DigitalNoiseInstance(noise_type, error_probability=1.1).kraus


def test_error_paulichannel():
    with pytest.raises(ValueError):
        noise = DigitalNoiseInstance(
            DigitalNoiseType.PAULI_CHANNEL, error_probability=(0.4, 0.5, 1.1)
        ).kraus

    for p in range(3):
        probas = [1.0 / 3.0] * 3
        probas[p] = -0.1
        with pytest.raises(ValueError):
            noise = DigitalNoiseInstance(
                DigitalNoiseType.PAULI_CHANNEL, error_probability=probas
            ).kraus

        probas = [0.0] * 3
        probas[p] = 1.1
        with pytest.raises(ValueError):
            noise = DigitalNoiseInstance(
                DigitalNoiseType.PAULI_CHANNEL, error_probability=probas
            ).kraus


def test_error_prob_GeneralizedAmplitudeDamping():
    for p in range(2):
        probas = [1.0 / 2.0] * 2
        probas[p] = -0.1
        with pytest.raises(ValueError):
            noise = DigitalNoiseInstance(
                DigitalNoiseType.GENERALIZED_AMPLITUDE_DAMPING, error_probability=probas
            ).kraus

        probas = [0.0] * 2
        probas[p] = 1.1
        with pytest.raises(ValueError):
            noise = DigitalNoiseInstance(
                DigitalNoiseType.GENERALIZED_AMPLITUDE_DAMPING, error_probability=probas
            ).kraus


@pytest.mark.parametrize("gate_fn", PRIMITIVE_GATES)
@pytest.mark.parametrize("noise_type", ALL_NOISES)
@pytest.mark.parametrize("zero_proba", [False, True])
def test_noisy_primitive(gate_fn: Callable, noise_type: DigitalNoiseType, zero_proba: bool) -> None:
    target = np.random.randint(0, MAX_QUBITS)
    noise = noise_instance(noise_type, zero_proba)

    noisy_gate = gate_fn(target, noise=(noise,))
    assert len(noisy_gate.noise) == 1

    dm_shape_len = 2 * MAX_QUBITS

    orig_state = random_state(MAX_QUBITS)
    output_dm = apply_gates(orig_state, noisy_gate)

    # check output is a density matrix
    assert len(output_dm.array.shape) == dm_shape_len

    orig_dm = density_mat(orig_state)
    assert len(orig_dm.array.shape) == dm_shape_len
    output_dm2 = apply_gates(
        orig_dm,
        noisy_gate,
    )
    assert jnp.allclose(output_dm2.array, output_dm.array)

    perfect_gate = gate_fn(target)
    perfect_output = density_mat(apply_gates(orig_state, perfect_gate))
    if zero_proba:
        assert jnp.allclose(perfect_output.array, output_dm.array)
    else:
        assert not jnp.allclose(perfect_output.array, output_dm.array)


@pytest.mark.parametrize("gate_fn", PARAMETRIC_GATES)
@pytest.mark.parametrize("noise_type", ALL_NOISES)
@pytest.mark.parametrize("zero_proba", [False, True])
def test_noisy_parametric(
    gate_fn: Callable, noise_type: DigitalNoiseType, zero_proba: bool
) -> None:
    target = np.random.randint(0, MAX_QUBITS)
    noise = noise_instance(noise_type, zero_proba)
    noisy_gate = gate_fn("theta", target, noise=(noise,))
    values = {"theta": np.random.uniform(0.1, 2 * np.pi)}
    orig_state = random_state(MAX_QUBITS)

    dm_shape_len = 2 * MAX_QUBITS

    output_dm = apply_gates(orig_state, noisy_gate, values)
    # check output is a density matrix
    assert len(output_dm.array.shape) == dm_shape_len

    orig_dm = density_mat(orig_state)
    assert len(orig_dm.array.shape) == dm_shape_len

    output_dm2 = apply_gates(
        orig_dm,
        noisy_gate,
        values,
    )
    assert jnp.allclose(output_dm2.array, output_dm.array)

    perfect_gate = gate_fn("theta", target)
    perfect_output = density_mat(apply_gates(orig_state, perfect_gate, values))
    if zero_proba:
        assert jnp.allclose(perfect_output.array, output_dm.array)
    else:
        assert not jnp.allclose(perfect_output.array, output_dm.array)


def simple_depolarizing_test() -> None:
    noise = (DigitalNoiseInstance(DigitalNoiseType.DEPOLARIZING, 0.1),)
    ops = [X(0, noise=noise), X(1)]
    state = product_state("00")
    state_output = run(ops, state)

    # test run
    assert jnp.allclose(
        state_output,
        jnp.array(
            [
                [
                    [[0.0 - 0.0j, 0.0 - 0.0j], [0.0 - 0.0j, 0.0 - 0.0j]],
                    [[0.0 - 0.0j, 0.06666667 - 0.0j], [0.0 - 0.0j, 0.0 - 0.0j]],
                ],
                [
                    [[0.0 - 0.0j, 0.0 - 0.0j], [0.0 - 0.0j, 0.0 - 0.0j]],
                    [[0.0 - 0.0j, 0.0 - 0.0j], [0.0 - 0.0j, 0.93333333 - 0.0j]],
                ],
            ],
            dtype=jnp.complex128,
        ),
    )

    # test sampling
    dm_state = density_mat(state)
    sampling_output = sample(
        dm_state,
        ops,
    )
    assert "11" in sampling_output.keys()
    assert "01" in sampling_output.keys()

    # test expectation
    exp_dm = expectation(dm_state, ops, [Z(0)], {})
    assert jnp.allclose(exp_dm, jnp.array([-0.86666667], dtype=jnp.float64))

    # test shots expectation
    exp_dm_shots = expectation(dm_state, ops, [Z(0)], {}, n_shots=1000)
    assert jnp.allclose(exp_dm, exp_dm_shots, atol=1e-02)


def test_error_noisy_gate_sparse() -> None:
    noise_type = ALL_NOISES[0]
    noise = noise_instance(noise_type)

    noisy_gate = X(0, noise=(noise,), sparse=True)
    state = product_state("00", sparse=True)

    with pytest.raises(NotImplementedError):
        state_output = apply_gates(state, noisy_gate)



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import Any

import jax
import jax.numpy as jnp
from jax import Array
from jax.experimental.sparse import BCOO, bcoo_concatenate, empty
from jax.tree_util import register_pytree_node_class

from horqrux._misc import default_complex_dtype
from horqrux.noise import NoiseProtocol
from horqrux.utils.operator_utils import (
    ControlQubits,
    QubitSupport,
    TargetQubits,
    _jacobian,
    _unitary,
    is_controlled,
)
from horqrux.utils.sparse_utils import eigvalsh_sp

from .primitive import Primitive

default_dtype = default_complex_dtype()
nonzero_jit = jax.jit(jnp.nonzero, static_argnames="size")
unique_jit = jax.jit(jnp.unique, static_argnames="size")


@register_pytree_node_class
@dataclass
class Parametric(Primitive):
    """Extension of the Primitive class adding the option to pass a parameter."""

    generator_name: str
    target: QubitSupport
    control: QubitSupport
    noise: NoiseProtocol = None
    sparse: bool = False
    param: str | float = ""
    shift: float = 0.0

    def __post_init__(self) -> None:
        super().__post_init__()

        def parse_dict(values: dict[str, float] = dict()) -> float:
            # note: shift is for GPSR when the same param_name is used in many operations
            return values[self.param] + self.shift  # type: ignore[index]

        def parse_val(values: dict[str, float] = dict()) -> float:
            return self.param + self.shift  # type: ignore[return-value, operator]

        self.parse_values = parse_dict if isinstance(self.param, str) else parse_val

    def tree_flatten(  # type: ignore[override]
        self,
    ) -> tuple[tuple, tuple[str, tuple, tuple, NoiseProtocol, bool, str | float, float]]:
        children = ()
        aux_data = (
            self.generator_name,
            self.target[0],
            self.control[0],
            self.noise,
            self.sparse,
            self.param,
            self.shift,
        )
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)

    def _unitary(self, values: dict[str, float] = dict()) -> Array:
        return _unitary(self.generator, self.parse_values(values))

    def jacobian(self, values: dict[str, float] = dict()) -> Array:
        return _jacobian(self.generator, self.parse_values(values))

    @property
    def name(self) -> str:
        base_name = "R" + self.generator_name
        return "C" + base_name if is_controlled(self.control) else base_name

    def __repr__(self) -> str:
        return (
            self.name
            + f"(target={self.target}, control={self.control}, param={self.param}, shift={self.shift})"
        )

    @cached_property
    def eigenvals_generator(self) -> Array:
        """Get eigenvalues of the underlying operation.

        Arguments:
            values: Parameter values.

        Returns:
            Array: Eigenvalues of the operation.
        """
        eig_vals_generator = eigvalsh_sp(self.generator)
        if is_controlled(self.control):
            if not self.sparse:
                eig_vals_generator = jnp.concatenate(
                    (
                        jnp.zeros(2 ** (len(self.control)), dtype=eig_vals_generator.dtype),
                        eig_vals_generator,
                    )
                )
            else:
                eig_vals_generator = bcoo_concatenate(
                    empty(2 ** (len(self.control)), dtype=eig_vals_generator.dtype),
                    eig_vals_generator,
                )
        return eig_vals_generator

    @cached_property
    def spectral_gap(self) -> Array:
        """Difference between the moduli of the two largest eigenvalues of the generator.

        Returns:
            Array: Spectral gap value.
        """
        spectrum = jnp.atleast_2d(self.eigenvals_generator)
        diffs = spectrum - spectrum.T
        # note for jitting, must specify a size
        # atm only size 2 is acceptable given all possible generators in OPERATIONS_DICT
        spectral_gap = unique_jit(jnp.abs(jnp.tril(diffs)), size=2)
        return spectral_gap[nonzero_jit(spectral_gap, size=1)]

    @property
    def is_parametric(self) -> bool:
        return isinstance(self.param, str)


def RX(
    param: float | str,
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Parametric:
    """RX gate.

    Arguments:
        param: Parameter denoting the Rotational angle.
        target: tuple of target qubits denoted as ints.
        control: Optional tuple of control qubits denoted as ints.
        noise: The noise instance. Defaults to None.

    Returns:
        Parametric: A Parametric gate object.
    """
    return Parametric("X", target, control, noise, param=param, sparse=sparse)


def RY(
    param: float | str,
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Parametric:
    """RY gate.

    Arguments:
        param: Parameter denoting the Rotational angle.
        target: tuple of target qubits denoted as ints.
        control: Optional tuple of control qubits denoted as ints.
        noise: The noise instance. Defaults to None.

    Returns:
        Parametric: A Parametric gate object.
    """
    return Parametric("Y", target, control, noise, param=param, sparse=sparse)


def RZ(
    param: float | str,
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Parametric:
    """RZ gate.

    Arguments:
        param: Parameter denoting the Rotational angle.
        target: tuple of target qubits denoted as ints.
        control: Optional tuple of control qubits denoted as ints.
        noise: The noise instance. Defaults to None.

    Returns:
        Parametric: A Parametric gate object.
    """
    return Parametric("Z", target, control, noise, param=param, sparse=sparse)


class _PHASE(Parametric):
    def _unitary(self, values: dict[str, float] = dict()) -> Array:
        u = jnp.eye(2, 2, dtype=default_dtype)
        u = u.at[(1, 1)].set(jnp.exp(1.0j * self.parse_values(values)))
        if self.sparse:
            u = BCOO.fromdense(u)
        return u

    def jacobian(self, values: dict[str, float] = dict()) -> Array:
        jac = jnp.zeros((2, 2), dtype=default_dtype)
        jac = jac.at[(1, 1)].set(1j * jnp.exp(1.0j * self.parse_values(values)))
        if self.sparse:
            jac = BCOO.fromdense(jac)
        return jac

    @property
    def name(self) -> str:
        base_name = "PHASE"
        return "C" + base_name if is_controlled(self.control) else base_name


def PHASE(
    param: float,
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Parametric:
    """Phase gate.

    Arguments:
        param: Parameter denoting the Rotational angle.
        target: tuple of target qubits denoted as ints.
        control: Optional tuple of control qubits denoted as ints.
        noise: The noise instance. Defaults to None.

    Returns:
        Parametric: A Parametric gate object.
    """

    return _PHASE("I", target, control, noise, param=param, sparse=sparse)



from __future__ import annotations

import chex
import jax
import numpy as np
from absl.testing import parameterized
from jax import Array, value_and_grad

from horqrux import expectation, random_state
from horqrux.circuit import QuantumCircuit
from horqrux.composite import Observable
from horqrux.primitives.parametric import PHASE, RX, RY, RZ
from horqrux.primitives.primitive import NOT, H, I, S, T, X, Y, Z
from horqrux.utils.conversion import to_sparse
from horqrux.utils.operator_utils import DiffMode
from tests.utils import verify_arrays

MAX_QUBITS = 7
PARAMETRIC_GATES = (RX, RY, RZ, PHASE)
PRIMITIVE_GATES = (NOT, H, X, Y, Z, I, S, T)


class TestAdjoint(chex.TestCase):
    @chex.variants(with_jit=False, without_jit=True)
    @parameterized.parameters(True, False)
    def test_gradcheck(self, same_name: bool) -> None:
        names = ["theta", "epsilon", "phi", "omega"]
        if same_name:
            while len(set(names)) > 3:
                names = np.random.choice(names, len(names))
        ops = [
            RX(names[0], 0),
            RY(names[1], 0),
            RX(names[2], 0),
            NOT(1, 0),
            RX(names[3], 0, 1),
            NOT(1, 0),
            RY(1.0, 0),
        ]
        circuit = QuantumCircuit(2, ops)
        circuit_sparse = to_sparse(circuit)
        observable = [Observable([Z(0)])]
        observable_sparse = [to_sparse(observable[0])]
        values = {name: np.random.uniform(0, 1) for name in set(names)}
        state = random_state(MAX_QUBITS)
        state_sparse = random_state(MAX_QUBITS, sparse=True)

        @self.variant(static_argnums=(1,))
        def exp_fn(values: dict, diff_mode: DiffMode = "ad") -> Array:
            return expectation(state, circuit, observable, values, diff_mode).sum()

        @self.variant(static_argnums=(1,))
        def exp_fn_sparse(values: dict, diff_mode: DiffMode = "ad") -> Array:
            return expectation(
                state_sparse, circuit_sparse, observable_sparse, values, diff_mode
            ).sum()

        exp_ad, grad_ad = value_and_grad(exp_fn)(values)
        exp_adjoint, grads_adjoint = value_and_grad(exp_fn)(values, "adjoint")
        assert verify_arrays(exp_ad, exp_adjoint)

        # note grad does not work for sparse operations
        exp_adjoint_sparse = exp_fn_sparse(values, "adjoint")
        with self.assertRaises(NotImplementedError):
            jax.experimental.sparse.grad(exp_fn_sparse)(values, "adjoint")

        assert verify_arrays(exp_adjoint, exp_adjoint_sparse)

        for param, ad_grad in grad_ad.items():
            assert verify_arrays(grads_adjoint[param], ad_grad, atol=1.0e-3)



from __future__ import annotations

import jax.numpy as jnp
import numpy as np
import pytest
from jax import jit, vmap

from horqrux.analog import HamiltonianEvolution
from horqrux.apply import apply_gates
from horqrux.utils.operator_utils import is_normalized, overlap, random_state, uniform_state

sigmaz = jnp.diag(jnp.array([1.0, -1.0], dtype=jnp.cdouble))
Hbase = jnp.kron(sigmaz, sigmaz)

Hamiltonian = jnp.kron(Hbase, Hbase)


def test_hamevo_single() -> None:
    n_qubits = 4
    t_evo = jnp.pi / 4
    hamevo = HamiltonianEvolution(tuple([i for i in range(n_qubits)]))
    psi = uniform_state(n_qubits)
    psi_star = apply_gates(psi, hamevo, {"hamiltonian": Hamiltonian, "time_evolution": t_evo})
    result = overlap(psi_star, psi)
    assert jnp.isclose(result, 0.5)


def Hamiltonian_general(n_qubits: int = 2, batch_size: int = 1) -> jnp.array:
    H_batch = jnp.zeros((batch_size, 2**n_qubits, 2**n_qubits), dtype=jnp.cdouble)
    for i in range(batch_size):
        H_0 = np.random.uniform(0.0, 1.0, (2**n_qubits, 2**n_qubits)).astype(np.cdouble)
        H = H_0 + jnp.conj(H_0.transpose(0, 1))
        H_batch.at[i, :, :].set(H)
    return H_batch


@pytest.mark.parametrize(
    "n_qubits, batch_size",
    [
        (2, 1),
    ],
)
def test_hamevo_general(n_qubits: int, batch_size: int) -> None:
    H = Hamiltonian_general(n_qubits, batch_size)
    t_evo = np.random.uniform(0, 1, (batch_size, 1))
    hamevo = HamiltonianEvolution(tuple([i for i in range(n_qubits)]))
    psi = random_state(n_qubits)
    psi_star = jit(
        vmap(apply_gates, in_axes=(None, None, {"hamiltonian": 0, "time_evolution": 0}))
    )(psi, hamevo, {"hamiltonian": H, "time_evolution": t_evo})
    assert jnp.all(vmap(is_normalized, in_axes=(0,))(psi_star))



from __future__ import annotations

from functools import partial, reduce, singledispatch
from operator import add
from typing import Any, Iterable, Union

import jax
import jax.numpy as jnp
import numpy as np
from jax import Array
from jax.experimental.sparse import BCOO, sparsify

from horqrux.noise import NoiseProtocol
from horqrux.primitives.primitive import Primitive
from horqrux.utils.operator_utils import (
    DensityMatrix,
    OperationType,
    State,
    _controlled,
    _dagger,
    density_mat,
    expand_operator,
    is_controlled,
    num_qubits,
    permute_basis,
)


@singledispatch
def apply_operator(
    state: Any,
    operator: Array,
    target: tuple[int, ...],
    control: tuple[Union[int, None], ...],
) -> Any:
    """Apply an operator on a state or density matrix.

    Args:
        state (Any): Array to operate on.
        operator (Array): Array to contract over 'state'.
        target (tuple[int, ...]): tuple of target qubits on which to apply the 'operator' to.
        control (tuple[int  |  None, ...]): tuple of control qubits.

    Raises:
        NotImplementedError: If not implemented for given types.

    Returns:
        Array: The output of the application of the operator.
    """
    raise NotImplementedError("apply_operator is not implemented")


@apply_operator.register
def _(
    state: Array,
    operator: Array,
    target: tuple[int, ...],
    control: tuple[Union[int, None], ...],
) -> Array:
    """Applies an operator, i.e. a single array of shape [2, 2, ...], on a given state
       of shape [2 for _ in range(n_qubits)] for a given set of target and control qubits.
       In case of a controlled operation, the 'operator' array will be embedded into a controlled array.

       Since dimension 'i' in 'state' corresponds to all amplitudes where qubit 'i' is 1,
       target and control qubits represent the dimensions over which to contract the 'operator'.
       Contraction means applying the 'dot' operation between the operator array and dimension 'i'
       of 'state, resulting in a new state where the result of the 'dot' operation has been moved to
       dimension 'i' of 'state'. To restore the former order of dimensions, the affected dimensions
       are moved to their original positions and the state is returned.

    Args:
        state (Array): Array to operate on.
        operator (Array): Array to contract over 'state'.
        target (tuple[int, ...]): tuple of target qubits on which to apply the 'operator' to.
        control (tuple[int  |  None, ...]): tuple of control qubits.

    Returns:
        Array after applying 'operator'.
    """
    state_dims: tuple[int, ...] = target
    if is_controlled(control):
        operator = _controlled(operator, len(control))
        state_dims = (*control, *target)  # type: ignore[arg-type]
    n_qubits_op = int(np.log2(operator.shape[1]))
    operator = operator.reshape((2,) * (2 * n_qubits_op))
    op_out_dims = tuple(np.arange(operator.ndim // 2, operator.ndim, dtype=int))
    # Apply operator
    new_state_dims = tuple(range(len(state_dims)))
    state = jnp.tensordot(a=operator, b=state, axes=(op_out_dims, state_dims))
    return jnp.moveaxis(a=state, source=new_state_dims, destination=state_dims)


@apply_operator.register
def _(
    state: BCOO,
    operator: BCOO,
    target: tuple[int, ...],
    control: tuple[Union[int, None], ...],
) -> Array:
    """Applies an operator, i.e. a single array of shape [2, 2, ...], on a given state
       of shape [2] * n_qubits for a given set of target and control qubits.
       In case of a controlled operation, the 'operator' array will be embedded into a controlled array.

       Since dimension 'i' in 'state' corresponds to all amplitudes where qubit 'i' is 1,
       target and control qubits represent the dimensions over which to contract the 'operator'.
       Contraction means applying the 'dot' operation between the operator array and dimension 'i'
       of 'state, resulting in a new state where the result of the 'dot' operation has been moved to
       dimension 'i' of 'state'. To restore the former order of dimensions, the affected dimensions
       are moved to their original positions and the state is returned.

    Args:
        state (Array): Array to operate on.
        operator (Array): Array to contract over 'state'.
        target (tuple[int, ...]): tuple of target qubits on which to apply the 'operator' to.
        control (tuple[int  |  None, ...]): tuple of control qubits.

    Returns:
        Array after applying 'operator'.
    """
    state_dims: tuple[int, ...] = target
    if is_controlled(control):
        operator = _controlled(operator, len(control))
        state_dims = (*control, *target)  # type: ignore[arg-type]
    n_qubits_op = int(np.log2(operator.shape[1]))
    operator = operator.reshape((2,) * (2 * n_qubits_op))
    op_out_dims = tuple(np.arange(operator.ndim // 2, operator.ndim, dtype=int))
    # Apply operator
    new_state_dims = tuple(range(len(state_dims)))
    tensordot_sp = sparsify(lambda a, b: jnp.tensordot(a=a, b=b, axes=(op_out_dims, state_dims)))
    moveaxis_sp = sparsify(
        lambda a: jnp.moveaxis(a=a, source=new_state_dims, destination=state_dims)
    )
    state = tensordot_sp(operator, state)
    return moveaxis_sp(state)


@apply_operator.register
def _(
    state: DensityMatrix,
    operator: Array,
    target: tuple[int, ...],
    control: tuple[Union[int, None], ...],
) -> DensityMatrix:
    """Applies an operator, i.e. a single array of shape [2, 2, ...], on a given density matrix
       of shape [2 for _ in range(2 * n_qubits)] for a given set of target and control qubits.
       In case of a controlled operation, the 'operator' array will be embedded into a controlled array.

    Args:
        state (DensityMatrix): Array to operate on.
        operator (Array): Array to contract over 'state'.
        target (tuple[int, ...]): tuple of target qubits on which to apply the 'operator' to.
        control (tuple[int  |  None, ...]): tuple of control qubits.

    Returns:
        Density matrix after applying 'operator'.
    """
    state_dims: tuple[int, ...] = target
    if is_controlled(control):
        operator = _controlled(operator, len(control))
        state_dims = (*control, *target)  # type: ignore[arg-type]
    n_qubits_op = int(np.log2(operator.shape[1]))
    operator = operator.reshape((2,) * (2 * n_qubits_op))
    op_out_dims = tuple(np.arange(operator.ndim // 2, operator.ndim, dtype=int))
    op_in_dims = tuple(np.arange(0, operator.ndim // 2, dtype=int))
    new_state_dims = tuple(range(len(state_dims)))

    # Apply operator to density matrix: ρ' = O ρ O†
    out_state = state.array
    support_perm = state_dims + tuple(set(tuple(range(out_state.ndim // 2))) - set(state_dims))

    out_state = permute_basis(out_state, support_perm, False)
    out_state = jnp.tensordot(a=operator, b=out_state, axes=(op_out_dims, new_state_dims))

    out_state = _dagger(out_state)
    out_state = jnp.tensordot(a=operator, b=out_state, axes=(op_out_dims, op_in_dims))
    out_state = _dagger(out_state)

    out_state = permute_basis(out_state, support_perm, True)
    return DensityMatrix(out_state)


def apply_kraus_operator(
    kraus: Array,
    array: Array,
    target: tuple[int, ...],
) -> Array:
    """Apply K \\rho K^\\dagger.

    Args:
        kraus (Array): Kraus operator K.
        state (Array): Input density matrix.
        target (tuple[int, ...]): Target qubits.

    Returns:
        Array: K \\rho K^\\dagger.
    """
    state_dims: tuple[int, ...] = target
    n_qubits = int(np.log2(kraus.size))
    kraus = kraus.reshape((2,) * n_qubits)
    op_out_dims = tuple(np.arange(kraus.ndim // 2, kraus.ndim, dtype=int))
    op_in_dims = tuple(np.arange(0, kraus.ndim // 2, dtype=int))
    new_state_dims = tuple(range(len(state_dims)))

    array = jnp.moveaxis(a=array, source=new_state_dims, destination=state_dims)
    array = jnp.tensordot(a=kraus, b=array, axes=(op_out_dims, new_state_dims))

    array = _dagger(array)
    array = jnp.tensordot(a=kraus, b=array, axes=(op_out_dims, op_in_dims))
    array = _dagger(array)
    array = jnp.moveaxis(a=array, source=state_dims, destination=new_state_dims)
    return array


def apply_kraus_sum(
    kraus_ops: Array,
    array: Array,
    target: tuple[int, ...],
) -> DensityMatrix:
    """Apply the following evolution as a sum of Kraus operators:
        .. math::
            S(\\rho) = \\sum_i K_i \\rho K_i^\\dagger

    Args:
        kraus_ops (Array): Stacked K_i.
        state (Array): Input array.
        target (tuple[int, ...]): Qubits the operator is defined on.

    Returns:
        DensityMatrix: Output density matrix.
    """
    apply_one_kraus = jax.vmap(
        partial(
            apply_kraus_operator,
            array=array,
            target=target,
        )
    )
    kraus_evol = apply_one_kraus(kraus_ops)
    output_dm = jnp.sum(kraus_evol, 0)
    return DensityMatrix(output_dm)


def apply_operator_with_noise(
    state: DensityMatrix,
    operator: Array,
    target: tuple[int, ...],
    control: tuple[Union[int, None], ...],
    noise: NoiseProtocol,
) -> State:
    """Evolves the input state and applies a noisy quantum channel
       on the evolved state :math:`\rho`.

        The evolution is represented as a sum of Kraus operators:
        .. math::
            S(\\rho) = \\sum_i K_i \\rho K_i^\\dagger,

    Args:
        state (State): Input state or density matrix.
        operator (Array): Operator to apply.
        target (tuple[int, ...]): Target qubits.
        control (tuple[int  |  None, ...]): Control qubits.
        noise (NoiseProtocol): The noise protocol.

    Returns:
        Array: Output state or density matrix.
    """
    state_gate = apply_operator(state, operator, target, control)
    if noise is None:
        return state_gate
    else:
        output_dm = state_gate
        full_support = tuple(range(num_qubits(state_gate)))
        for n in noise:
            kraus_ops = jnp.stack(list(expand_operator(k, target, full_support) for k in n.kraus))
            output_dm = apply_kraus_sum(kraus_ops, output_dm.array, full_support)
        return output_dm


def group_by_index(gates: Iterable[Primitive]) -> Iterable[Primitive]:
    """Group gates together which are acting on the same qubit."""
    sorted_gates = []
    gate_batch = []
    for gate in gates:
        if not is_controlled(gate.control):
            gate_batch.append(gate)
        else:
            if len(gate_batch) > 0:
                gate_batch.sort(key=lambda g: g.target)
                sorted_gates += gate_batch
                gate_batch = []
            sorted_gates.append(gate)
    if len(gate_batch) > 0:
        gate_batch.sort(key=lambda g: g.target)
        sorted_gates += gate_batch
    return sorted_gates


def merge_operators(
    operators: tuple[Array, ...], targets: tuple[int, ...], controls: tuple[int, ...]
) -> tuple[tuple[Array, ...], tuple[int, ...], tuple[int, ...]]:
    """
    If possible, merge several operators acting on the same qubits into a single array
    which can then be contracted over a state in a single tensordot operation.

    Arguments:
        operators: The arrays representing the unitaries to be merged.
        targets: The corresponding target qubits.
        controls: The corresponding control qubits.

    Returns:
        A tuple of merged operators, targets and controls.
    """
    if len(operators) < 2:
        return operators, targets, controls
    operators, targets, controls = operators[::-1], targets[::-1], controls[::-1]
    merged_operator, merged_target, merged_control = operators[0], targets[0], controls[0]
    merged_operators = merged_targets = merged_controls = tuple()  # type: ignore[var-annotated]
    for operator, target, control in zip(operators[1:], targets[1:], controls[1:]):
        if target == merged_target and control == merged_control:
            merged_operator = merged_operator @ operator
        else:
            merged_operators += (merged_operator,)
            merged_targets += (merged_target,)
            merged_controls += (merged_control,)
            merged_operator, merged_target, merged_control = operator, target, control
    if merged_operator is not None:
        merged_operators += (merged_operator,)
        merged_targets += (merged_target,)
        merged_controls += (merged_control,)
    return merged_operators[::-1], merged_targets[::-1], merged_controls[::-1]


@singledispatch
def apply_gates(
    state: Any,
    gate: Union[Primitive, Iterable[Primitive]],
    values: dict[str, float] = dict(),
    op_type: OperationType = OperationType.UNITARY,
    group_gates: bool = False,  # Defaulting to False since this can be performed once before circuit execution
    merge_ops: bool = True,
) -> Any:
    raise NotImplementedError("apply_gate is not implemented")


def prepare_sequence_reduce(
    gate: Union[Primitive, Iterable[Primitive]],
    values: dict[str, float] = dict(),
    op_type: OperationType = OperationType.UNITARY,
    group_gates: bool = False,  # Defaulting to False since this can be performed once before circuit execution
    merge_ops: bool = True,
) -> tuple[tuple[Array, ...], tuple, tuple, list[NoiseProtocol]]:
    """Prepare the tuples to be used when applying operations.

    Args:
        gate (Union[Primitive, Iterable[Primitive]]): Gate(s) to apply.
        values (dict[str, float], optional): A dictionary with parameter values.
            Defaults to dict().
        op_type (OperationType, optional): The type of operation to perform: Unitary, Dagger or Jacobian.
            Defaults to OperationType.UNITARY.
        group_gates (bool, optional): Group gates together which are acting on the same qubit.
            Defaults to False.

    Returns:
        tuple[tuple[Array, ...], tuple, tuple, list[NoiseProtocol]]: Operators, targets,
            controls and noise.
    """
    operator: tuple[Array, ...]
    noise = list()
    if isinstance(gate, Primitive):
        operator_fn = getattr(gate, op_type)
        operator, target, control = (operator_fn(values),), gate.target, gate.control
        noise += [gate.noise]
    else:
        if group_gates:
            gate = group_by_index(gate)
        operator = tuple(getattr(g, op_type)(values) for g in gate)
        target = reduce(add, [g.target for g in gate])
        control = reduce(add, [g.control for g in gate])
        noise = [g.noise for g in gate]
        # merge when no noise is present
        if (noise == [None] * len(noise)) and merge_ops:
            operator, target, control = merge_operators(operator, target, control)

    return operator, target, control, noise


@apply_gates.register
def _(
    state: Array,
    gate: Union[Primitive, Iterable[Primitive]],
    values: dict[str, float] = dict(),
    op_type: OperationType = OperationType.UNITARY,
    group_gates: bool = False,  # Defaulting to False since this can be performed once before circuit execution
    merge_ops: bool = True,
) -> State:
    """Wrapper function for 'apply_operator' which applies a gate or a series of gates to a given state.
    Arguments:
        state: Array or DensityMatrix to operate on.
        gate: Gate(s) to apply.
        values: A dictionary with parameter values.
        op_type: The type of operation to perform: Unitary, Dagger or Jacobian.
        group_gates: Group gates together which are acting on the same qubit.
        merge_ops: Attempt to merge operators acting on the same qubit.

    Returns:
        Array or density matrix after applying 'gate'.
    """
    operator, target, control, noise = prepare_sequence_reduce(
        gate, values, op_type, group_gates, merge_ops
    )

    # faster way to check has_noise
    has_noise = noise != [None] * len(noise)
    if has_noise:
        state = density_mat(state)
        output_state = reduce(
            lambda state, gate: apply_operator_with_noise(state, *gate),
            zip(operator, target, control, noise),
            state,
        )
    else:
        output_state = reduce(
            lambda state, gate: apply_operator(state, *gate),
            zip(operator, target, control),
            state,
        )
    return output_state


@apply_gates.register
def _(
    state: BCOO,
    gate: Union[Primitive, Iterable[Primitive]],
    values: dict[str, float] = dict(),
    op_type: OperationType = OperationType.UNITARY,
    group_gates: bool = False,  # Defaulting to False since this can be performed once before circuit execution
    merge_ops: bool = True,
) -> State:
    """Wrapper function for 'apply_operator' which applies a gate or a series of gates to a given state.
    Arguments:
        state: Array or DensityMatrix to operate on.
        gate: Gate(s) to apply.
        values: A dictionary with parameter values.
        op_type: The type of operation to perform: Unitary, Dagger or Jacobian.
        group_gates: Group gates together which are acting on the same qubit.
        merge_ops: Attempt to merge operators acting on the same qubit.

    Returns:
        Array or density matrix after applying 'gate'.
    """
    operator, target, control, noise = prepare_sequence_reduce(
        gate, values, op_type, group_gates, merge_ops
    )

    # faster way to check has_noise
    has_noise = noise != [None] * len(noise)
    if has_noise:
        raise NotImplementedError("Noisy simulations are not supported with sparse operators.")
    else:
        output_state = reduce(
            lambda state, gate: apply_operator(state, *gate),
            zip(operator, target, control),
            state,
        )
    return output_state


@apply_gates.register
def _(
    state: DensityMatrix,
    gate: Union[Primitive, Iterable[Primitive]],
    values: dict[str, float] = dict(),
    op_type: OperationType = OperationType.UNITARY,
    group_gates: bool = False,  # Defaulting to False since this can be performed once before circuit execution
    merge_ops: bool = True,
) -> DensityMatrix:
    """Wrapper function for 'apply_operator' which applies a gate or a series of gates to a given state.
    Arguments:
        state: Array or DensityMatrix to operate on.
        gate: Gate(s) to apply.
        values: A dictionary with parameter values.
        op_type: The type of operation to perform: Unitary, Dagger or Jacobian.
        group_gates: Group gates together which are acting on the same qubit.
        merge_ops: Attempt to merge operators acting on the same qubit.

    Returns:
        Array or density matrix after applying 'gate'.
    """
    operator, target, control, noise = prepare_sequence_reduce(
        gate, values, op_type, group_gates, merge_ops
    )
    output_state = reduce(
        lambda state, gate: apply_operator_with_noise(state, *gate),
        zip(operator, target, control, noise),
        state,
    )
    return output_state



from __future__ import annotations

import chex
import jax
from jax import Array

from horqrux import expectation, random_state
from horqrux.circuit import QuantumCircuit
from horqrux.composite import Observable
from horqrux.primitives.parametric import RX, RZ
from tests.utils import verify_arrays


class ADTest(chex.TestCase):
    @chex.variants(with_jit=True, without_jit=True)
    def test_ad_obs(self) -> None:
        param_name = "theta"
        x = jax.random.uniform(jax.random.key(0), (3,))
        param_names = [param_name, param_name + "2", param_name + "obs"]
        ops = [RX(param_names[0], 0), RX(param_names[1], 1)]

        def values_to_dict(x, y=None) -> dict[str, Array]:
            values_circuit = {param_names[0]: x[0], param_names[1]: x[1], param_names[2]: x[2]}
            if y is None:
                return values_circuit
            else:
                return {"circuit": values_circuit, "observables": {param_names[2]: y}}

        circuit = QuantumCircuit(2, ops)
        observables = [Observable([RZ(param_name + "obs", 0)])]
        state = random_state(2)

        @self.variant
        def exp_fn(x, y=None) -> Array:
            values = values_to_dict(x, y)
            return expectation(state, circuit, observables, values, diff_mode="ad")

        d_exact = jax.grad(lambda x: exp_fn(x).sum())(x)

        d_exact_circuit_params = jax.grad(lambda x, y: exp_fn(x, y).sum(), argnums=0)(x[:2], x[-1])
        d_exact_obs_params = jax.grad(lambda x, y: exp_fn(x, y).sum(), argnums=1)(x[:2], x[-1])

        assert verify_arrays(d_exact[:-1], d_exact_circuit_params)
        assert verify_arrays(d_exact[-1], d_exact_obs_params)



from __future__ import annotations

from functools import singledispatch
from typing import Any, Iterable, Union

import jax
import jax.numpy as jnp
from jax import Array
from jax.experimental import checkify

from horqrux.apply import apply_gates
from horqrux.composite import Observable
from horqrux.primitives import Primitive
from horqrux.utils.operator_utils import DensityMatrix, State, expand_operator, num_qubits


@singledispatch
def eigen_probabilities(state: Any, eigvecs: Array) -> Array:
    """Obtain the probabilities using an input state and the eigenvectors decomposition
       of an observable.

    Args:
        state (Any): Input.
        eigvecs (Array): Eigenvectors of the observables.

    Returns:
        Array: The probabilities.
    """
    raise NotImplementedError(
        f"eigen_probabilities is not implemented for the state type {type(state)}."
    )


@eigen_probabilities.register
def _(state: Array, eigvecs: Array) -> Array:
    """Obtain the probabilities using an input quantum state vector
        and the eigenvectors decomposition
        of an observable.

    Args:
        state (Array): Input array.
        eigvecs (Array): Eigenvectors of the observables.

    Returns:
        Array: The probabilities.
    """
    inner_prod = jnp.matmul(jnp.conjugate(eigvecs.T), state.flatten())
    return jnp.abs(inner_prod) ** 2


@eigen_probabilities.register
def _(state: DensityMatrix, eigvecs: Array) -> Array:
    """Obtain the probabilities using an input quantum density matrix
        and the eigenvectors decomposition
        of an observable.

    Args:
        state (DensityMatrix): Input density matrix.
        eigvecs (Array): Eigenvectors of the observables.

    Returns:
        Array: The probabilities.
    """
    mat_prob = jnp.conjugate(eigvecs.T) @ state.array @ eigvecs
    return mat_prob.diagonal().real


def eigen_sample(
    state: State,
    observables: list[Observable],
    values: dict[str, float],
    n_qubits: int,
    n_shots: int,
    key: Any = jax.random.PRNGKey(0),
) -> Array:
    """Sample eigenvalues of observable given the probability distribution
        defined by applying the eigenvectors to the state.

    Args:
        state (State): Input state or density matrix.
        observables (list[Observable]): list of observables.
        values (dict[str, float]): Parameter values.
        n_qubits (int): Number of qubits
        n_shots (int): Number of samples
        key (Any, optional): Random seed key. Defaults to jax.random.PRNGKey(0).

    Returns:
        Array: Sampled eigenvalues.
    """
    qubits = tuple(range(n_qubits))
    d = 2**n_qubits
    mat_obs = list(
        map(
            lambda observable: expand_operator(
                observable.tensor(values), observable.qubit_support, qubits
            ).reshape((d, d)),
            observables,
        )
    )
    eigs = jax.vmap(jnp.linalg.eigh)(jnp.stack(mat_obs))
    eigvecs, eigvals = align_eigenvectors(eigs.eigenvalues, eigs.eigenvectors)
    probs = eigen_probabilities(state, eigvecs)
    return jax.random.choice(key=key, a=eigvals, p=probs, shape=(n_shots,)).mean(axis=0)


@jax.vmap
def validate_permutation_matrix(P: Array) -> Array:
    rows = P.sum(axis=0)
    columns = P.sum(axis=1)
    ones = jnp.ones(P.shape[0], dtype=rows.dtype)
    return ((ones == rows) & (ones == columns)).min()


def checkify_valid_permutation(P: Array) -> None:
    checkify.check(
        jnp.all(validate_permutation_matrix(P)),
        "Did not calculate valid permutation matrix",
    )


def align_eigenvectors(eigenvalues: Array, eigenvectors: Array) -> tuple[Array, Array]:
    """
    Given a list of eigenvalue eigenvector matrix tuples in the form of
    [(eigenvalue, eigenvector)...], this function aligns all the eigenvector
    matrices so that they are identical, and also rearranges the corresponding
    eigenvalues.

    This is primarily used as a utility function to help sample multiple
    correlated observables when using finite shots.
    """
    eigenvector = eigenvectors[0]

    P = jax.vmap(lambda mat: permutation_matrix(mat, eigenvector))(eigenvectors)
    checkify.checkify(checkify_valid_permutation)(P)
    aligned_eigenvalues = jax.vmap(jnp.dot)(eigenvalues, P).T
    return eigenvector, aligned_eigenvalues


def permutation_matrix(mat: Array, eigenvector_matrix: Array) -> Array:
    """Obtain the permutation matrix for aligning eigenvectors.

    Given two permuted eigenvector matrices, A and B, we wish to find a permutation
    matrix P such that A P = B. This function calculates such a permutation
    matrix and uses it to align each eigenvector matrix to the first eigenvector
    matrix of eigs.

    Args:
        mat (Array): Matrix A.
        eigenvector_matrix (Array): Eigenvector matrix B.

    Returns:
        Array: Permutation matrix P.
    """
    return (jnp.linalg.inv(mat) @ eigenvector_matrix).real > 0.5


def finite_shots(
    state: State,
    gates: Union[Primitive, Iterable[Primitive]],
    observables: list[Observable],
    values: dict[str, float],
    n_shots: int = 100,
    key: Any = jax.random.PRNGKey(0),
) -> Array:
    """Run 'state' through a sequence of 'gates' given parameters 'values'
    and compute the expectations using `n_shots` shots per observable.

    Args:
        state (State): Input state or density matrix.
        gates (Union[Primitive, Iterable[Primitive]]): Sequence of gates.
        observables (list[Observable]): List of observables.
        values (dict[str, float]): Parameter values.
        n_shots (int, optional): Number of shots. Defaults to 100.
        key (Any, optional): Key for randomness. Defaults to jax.random.PRNGKey(0).

    Returns:
        Array: Expectation values.
    """
    output_gates = apply_gates(state, gates, values)
    n_qubits = num_qubits(output_gates)
    if isinstance(output_gates, DensityMatrix):
        d = 2**n_qubits
        output_gates.array = output_gates.array.reshape((d, d))
    return eigen_sample(output_gates, observables, values, n_qubits, n_shots, key)



from __future__ import annotations

from .api import expectation, run, sample
from .apply import apply_gates, apply_operator
from .circuit import QuantumCircuit
from .composite import Add, Observable, Scale
from .primitives.parametric import PHASE, RX, RY, RZ
from .primitives.primitive import NOT, SWAP, H, I, S, T, X, Y, Z
from .utils.operator_utils import (
    DiffMode,
    equivalent_state,
    hilbert_reshape,
    overlap,
    product_state,
    random_state,
    uniform_state,
    zero_state,
)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Iterable

import numpy as np
from jax import Array
from jax.tree_util import register_pytree_node_class

from horqrux.matrices import OPERATIONS_DICT, SPARSE_OPERATIONS_DICT
from horqrux.noise import NoiseProtocol
from horqrux.utils.operator_utils import (
    ControlQubits,
    QubitSupport,
    State,
    TargetQubits,
    _dagger,
    controlled,
    expand_operator,
    is_controlled,
    none_like,
    zero_state,
)


@register_pytree_node_class
@dataclass
class Primitive:
    """Primitive gate class which stores information about generators target and control qubits
    of a particular quantum operator."""

    generator_name: str
    target: QubitSupport
    control: QubitSupport
    noise: NoiseProtocol = None
    sparse: bool = False

    @staticmethod
    def parse_idx(
        idx: tuple,
    ) -> tuple:
        if isinstance(idx, (int, np.int64)):
            return ((idx,),)
        elif isinstance(idx, tuple):
            return (idx,)
        else:
            return (idx.astype(int),)

    def __post_init__(self) -> None:
        self.target = Primitive.parse_idx(self.target)
        if self.control is None:
            self.control = none_like(self.target)
        else:
            self.control = Primitive.parse_idx(self.control)

    def __call__(self, state: State | None = None, values: dict[str, Array] = dict()) -> State:
        from horqrux.apply import apply_gates

        if state is None:
            state = zero_state(len(self.qubit_support))
        return apply_gates(state, self, values)

    def __iter__(self) -> Iterable[Primitive]:
        return iter((self,))

    def tree_flatten(
        self,
    ) -> tuple[tuple, tuple[str, TargetQubits, ControlQubits, NoiseProtocol, bool]]:
        children = ()
        aux_data = (self.generator_name, self.target[0], self.control[0], self.noise, self.sparse)
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)

    @property
    def generator(self) -> Array:
        return (
            SPARSE_OPERATIONS_DICT[self.generator_name]
            if self.sparse
            else OPERATIONS_DICT[self.generator_name]
        )

    def _unitary(self, values: dict[str, float] = dict()) -> Array:
        """Obtain the base unitary from `generator_name`.

        Args:
            values (dict[str, float], optional): Parameter values. Defaults to dict().

        Returns:
            Array: The base unitary from `generator_name`.
        """
        return self.generator

    def dagger(self, values: dict[str, float] = dict()) -> Array:
        """Obtain the dagger of the base unitary from `generator_name`.

        Args:
            values (dict[str, float], optional): Parameter values. Defaults to dict().

        Returns:
            Array: The base unitary daggered from `generator_name`.
        """
        return _dagger(self._unitary(values))

    def tensor(
        self,
        values: dict[str, float] = dict(),
        full_support: tuple[int, ...] | None = None,
    ) -> Array:
        """Obtain the unitary taking into account the qubit support for controlled operations.

        Args:
            values (dict[str, float], optional): Parameter values. Defaults to dict().
            full_support (tuple[int, ...], optional): The qubit support of definition for the unitary.

        Returns:
            Array: Unitary representation taking into account the qubit support.
        """
        base_unitary = self._unitary(values)
        if is_controlled(self.control):
            base_unitary = controlled(base_unitary, self.target, self.control)
        if full_support is None:
            return base_unitary
        else:
            return expand_operator(base_unitary, self.qubit_support, full_support)

    @property
    def name(self) -> str:
        return "C" + self.generator_name if is_controlled(self.control) else self.generator_name

    @property
    def n_qubits(self) -> int:
        n_qubits = len(self.target)
        if self.control[0] is not None:
            n_qubits += len(self.control)
        return n_qubits

    @property
    def qubit_support(self) -> tuple:
        return tuple(
            sorted(
                tuple(
                    set(
                        self.target[0] + self.control[0]
                        if is_controlled(self.control)
                        else self.target[0]
                    )
                )
            )
        )

    @property
    def is_parametric(self) -> bool:
        return False

    def __repr__(self) -> str:
        return self.name + f"(target={self.target}, control={self.control})"


def I(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    """Identity / I gate. This function returns an instance of 'Primitive' and does *not* apply the gate.
    By providing tuple of ints to 'control', it turns into a controlled gate.

    Example usage: I(1) represents the instruction to apply I to qubit 1.

    Args:
        target: tuple of ints describing the qubits to apply to.
        control: Optional tuple of ints indicating the control qubits. Defaults to (None,).
        noise: The noise instance. Defaults to None.
        sparse: True to use sparse arrays.

    Returns:
        A Primitive instance.
    """
    return Primitive("I", target, control, noise, sparse)


def X(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    """The definition for the X gate.

    This function returns an instance of 'Primitive' and does *not* apply the gate.
    By providing tuple of ints to 'control', it turns into a controlled gate.

    Example usage: X(1) represents the instruction to apply X to qubit 1.
    Example usage controlled: X(1, 0) represents the instruction to apply CX / CNOT to qubit 1 with controlled qubit 0.

    Args:
        target: tuple of ints describing the qubits to apply to.
        control: Optional tuple of ints indicating the control qubits. Defaults to (None,).
        noise: The noise instance. Defaults to None.
        sparse: True to use sparse arrays.

    Returns:
        A Primitive instance.
    """
    return Primitive("X", target, control, noise, sparse)


NOT = X


def Y(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    """Y gate. This function returns an instance of 'Primitive' and does *not* apply the gate.
    By providing tuple of ints to 'control', it turns into a controlled gate.

    Example usage: Y(1) represents the instruction to apply X to qubit 1.
    Example usage controlled: Y(1, 0) represents the instruction to apply CY to qubit 1 with controlled qubit 0.

    Args:
        target: tuple of ints describing the qubits to apply to.
        control: Optional tuple of ints indicating the control qubits. Defaults to (None,).
        noise: The noise instance. Defaults to None.
        sparse: True to use sparse arrays.

    Returns:
        A Primitive instance.
    """
    return Primitive("Y", target, control, noise, sparse)


def Z(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    """Z gate. This function returns an instance of 'Primitive' and does *not* apply the gate.
    By providing tuple of ints to 'control', it turns into a controlled gate.

    Example usage: Z(1) represents the instruction to apply Z to qubit 1.
    Example usage controlled: Z(1, 0) represents the instruction to apply CZ to qubit 1 with controlled qubit 0.

    Args:
        target: tuple of ints describing the qubits to apply to.
        control: Optional tuple of ints indicating the control qubits. Defaults to (None,).
        noise: The noise instance. Defaults to None.
        sparse: True to use sparse arrays.

    Returns:
        A Primitive instance.
    """
    return Primitive("Z", target, control, noise, sparse)


def H(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    """H/ Hadamard gate. This function returns an instance of 'Primitive' and does *not* apply the gate.
    By providing tuple of ints to 'control', it turns into a controlled gate.

    Example usage: H(1) represents the instruction to apply Hadamard to qubit 1.

    Args:
        target: tuple of ints describing the qubits to apply to.
        control: Optional tuple of ints indicating the control qubits. Defaults to (None,).
        noise: The noise instance. Defaults to None.
        sparse: True to use sparse arrays.

    Returns:
        A Primitive instance.
    """
    return Primitive("H", target, control, noise, sparse)


def S(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    """S gate or constant phase gate. This function returns an instance of 'Primitive' and does *not* apply the gate.
    By providing tuple of ints to 'control', it turns into a controlled gate.

    Example usage: S(1) represents the instruction to apply S to qubit 1.

    Args:
        target: tuple of ints describing the qubits to apply to.
        control: Optional tuple of ints indicating the control qubits. Defaults to (None,).
        noise: The noise instance. Defaults to None.
        sparse: True to use sparse arrays.

    Returns:
        A Primitive instance.
    """
    return Primitive("S", target, control, noise, sparse)


def T(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    """T gate. This function returns an instance of 'Primitive' and does *not* apply the gate.
    By providing tuple of ints to 'control', it turns into a controlled gate.

    Example usage: T(1) represents the instruction to apply Hadamard to qubit 1.

    Args:
        target: tuple of ints describing the qubits to apply to.
        control: Optional tuple of ints indicating the control qubits. Defaults to (None,).
        noise: The noise instance. Defaults to None.
        sparse: True to use sparse arrays.

    Returns:
        A Primitive instance.
    """
    return Primitive("T", target, control, noise, sparse)


# Multi (target) qubit gates


def SWAP(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    """SWAP gate. By providing a control, it turns into a controlled gate (Fredkin gate),
       use None for no control qubits.

    Example usage: SWAP((0, 1), (None,)) swaps qubits 0 and 1.
    Example usage controlled: SWAP(((0, 1), ), ((2, ))) swaps qubits 0 and 1 with controlled bit 2.

    Args:
        target: tuple of ints describing the qubits to apply to.
        control: Optional tuple of ints or None describing
        the control qubits. Defaults to (None,).
        noise: The noise instance. Defaults to None.
        sparse: True to use sparse arrays.

    Returns:
        A Primitive instance.
    """
    return Primitive("SWAP", target, control, noise, sparse)


def SQRTSWAP(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    return Primitive("SQSWAP", target, control, noise, sparse)


def ISWAP(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    return Primitive("ISWAP", target, control, noise, sparse)


def ISQRTSWAP(
    target: TargetQubits,
    control: ControlQubits = (None,),
    noise: NoiseProtocol = None,
    sparse: bool = False,
) -> Primitive:
    return Primitive("ISQSWAP", target, control, noise, sparse)



from __future__ import annotations

from functools import reduce
from operator import add
from typing import Any, Iterable

import jax.numpy as jnp
from jax import Array
from jax.tree_util import register_pytree_node_class

from horqrux.primitives import Primitive
from horqrux.utils.operator_utils import State

from .sequence import OpSequence


@register_pytree_node_class
class Scale(OpSequence):
    """
    Generic container for multiplying a 'Primitive', 'Sequence' or 'Add' instance by a parameter.

    Attributes:
        operations: Operations as a Sequence, Add, or a single Primitive operation.
        parameter_name: Name of the parameter to multiply operations with.
    """

    def __init__(
        self, operations: Primitive | OpSequence | list, parameter_name: str | float
    ) -> None:
        super().__init__(list(operations))  # type:ignore[arg-type]
        self.parameter: str | float = parameter_name

    def __call__(self, state: State | None = None, values: dict[str, Array] = dict()) -> Array:
        param = values[self.parameter] if isinstance(self.parameter, str) else self.parameter
        return jnp.array(param) * super().__call__(state, values)

    def tree_flatten(self) -> tuple:
        children = (self.operations,)
        aux_data = (self.parameter,)
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)

    def __iter__(self) -> Iterable[Scale]:
        return iter((self,))

    def tensor(
        self,
        values: dict[str, float] = dict(),
        full_support: tuple[int, ...] | None = None,
    ) -> Array:
        """Obtain the tensor representation.

        Args:
            values (dict[str, float], optional): Parameter values. Defaults to dict().
            full_support (tuple[int, ...], optional): The qubit support of definition for the unitary.

        Returns:
            Array: Unitary representation.
        """
        param = values[self.parameter] if isinstance(self.parameter, str) else self.parameter
        return param * super().tensor(values, full_support)


@register_pytree_node_class
class Add(OpSequence):
    """
    The 'add' operation applies all 'operations' to 'state' and returns the sum of states.

    Attributes:
        operations: List of operations to add up.
    """

    def __init__(self, operations: Primitive | OpSequence | list) -> None:
        super().__init__(list(operations))  # type:ignore[arg-type]

    def tree_flatten(self) -> tuple:
        children = (self.operations,)
        aux_data = ()
        return (children, aux_data)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*children, *aux_data)

    def __call__(self, state: State | None = None, values: dict[str, Array] = dict()) -> State:
        return reduce(add, map(lambda op: op(state, values), self.operations))

    def tensor(
        self,
        values: dict[str, float] = dict(),
        full_support: tuple[int, ...] | None = None,
    ) -> Array:
        """Obtain the unitary.

        Args:
            values (dict[str, float], optional): Parameter values. Defaults to dict().
            full_support (tuple[int, ...], optional): The qubit support of definition for the unitary.

        Returns:
            Array: Unitary representation.
        """
        if full_support is None:
            full_support = self.qubit_support
        elif not set(self.qubit_support).issubset(set(full_support)):
            raise ValueError(
                "Expanding tensor operation requires a `full_support` argument "
                "larger than or equal to the `qubit_support`."
            )
        return reduce(add, map(lambda op: op.tensor(values, full_support), self.operations))


@register_pytree_node_class
class Observable(Add):
    """
    The Observable :math:`O` represents an operator from which
    we can extract expectation values from quantum states.

    Given an input state :math:`\\ket\\rangle`, the expectation value with :math:`O` is defined as
    :math:`\\langle\\bra|O\\ket\\rangle`

    Attributes:
        operations: List of operations.
    """

    def __init__(self, operations: Primitive | OpSequence | list) -> None:
        super().__init__(list(operations))  # type:ignore[arg-type]

    def forward(self, state: State | None = None, values: dict[str, Array] = dict()) -> State:
        return super().__call__(state, values)

    def __call__(self, state: State | None = None, values: dict[str, Array] = dict()) -> Array:
        """Compute the expectation value using the observable.

        Args:
            state (State | None, optional): Input state. Defaults to None.
            values (dict[str, Array], optional): Parameter values. Defaults to dict().

        Returns:
            Array: Expectation values.
        """
        from horqrux.differentiation.ad import _ad_expectation_single_observable

        return _ad_expectation_single_observable(state, self, values)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable
from uuid import uuid4

from jax.tree_util import register_pytree_node_class

from horqrux.composite.sequence import OpSequence
from horqrux.primitives.parametric import RX, RY
from horqrux.primitives.primitive import NOT, Primitive


@register_pytree_node_class
@dataclass
class QuantumCircuit(OpSequence):
    """A minimalistic circuit class to store a sequence of gates.

    Attributes:
        n_qubits (int): Number of qubits.
        operations (list[Primitive]): Operations defining the circuit.
        fparams (list[str]): List of parameters that are considered
            non trainable, used for passing fixed input data to a quantum circuit.
                The corresponding operations compose the `feature map`.
    """

    def __init__(
        self, n_qubits: int, operations: Primitive | OpSequence | list, fparams: list[str] = list()
    ):
        super().__init__(list(operations))  # type:ignore[arg-type]
        self.n_qubits = n_qubits
        self.fparams = fparams

    @property
    def vparams(self) -> set[str]:
        """List of variational parameters of the circuit.

        Returns:
            list[str]: Names of variational parameters.
        """
        return set([name for name in self.param_names if name not in self.fparams])

    @property
    def n_vparams(self) -> int:
        """Number of variational parameters.

        Returns:
            int: Number of variational parameters.
        """
        return len(self.vparams)

    def tree_flatten(self) -> tuple:
        children = (
            self.operations,
            self.fparams,
        )
        aux_data = (self.n_qubits,)
        return (aux_data, children)

    @classmethod
    def tree_unflatten(cls, aux_data: Any, children: Any) -> Any:
        return cls(*aux_data, *children)


def hea(
    n_qubits: int,
    n_layers: int,
    rot_fns: list[Callable] = [RX, RY, RX],
    variational_param_prefix: str = "v_",
) -> list[Primitive]:
    """Hardware-efficient ansatz; A helper function to generate a sequence of rotations followed
    by a global entangling operation.

    Args:
        n_qubits (int): Number of qubits.
        n_layers (int): Number of layers
        rot_fns (list[Callable], optional): A list of rotations applied on one qubit.
            Defaults to [RX, RY, RX].
        variational_param_prefix (str, optional): Prefix for the name of variational parameters.
            Defaults to "v_". Names suffix are randomly generated strings with uuid4.

    Returns:
        list[Primitive]: List of gates composing the ansatz.
    """
    gates = []
    param_names = []
    for _ in range(n_layers):
        for i in range(n_qubits):
            ops = [
                fn(variational_param_prefix + str(uuid4()), qubit)
                for fn, qubit in zip(rot_fns, [i for _ in range(len(rot_fns))])
            ]
            param_names += [op.param for op in ops]
            ops += [NOT((i + 1) % n_qubits, i % n_qubits) for i in range(n_qubits)]  # type: ignore[arg-type]
            gates += ops

    return gates
